"""
Author: Mingjian He <mh1@stanford.edu>
"""

from .helper_functions import (
    np, plt,
    initialize_allosc, plot_mtm, plot_innovations,
    plot_residual, plot_acf, plot_pacf,
    get_innovations, get_knee, get_ar_psd,
    fit_ar, plot_fit_line, ebic_calc
    )
import pandas as pd
from ..utils import estimate_r


class DecomposedOscillatorModel(object):
    """
    DecomposedOscillatorModel is an object class containing fitted OscillatorModel objects
    from the decomposed oscillator search algorithm
    """
    def __init__(self, y, fs, noise_start=None, osc_range=7, iterate=False, **kwargs):
        """
        Inputs:
            :param y: observed data
            :param fs: sampling frequency (Hz)
            :param noise_start: frequency (Hz) above which white noise is assumed to estimate R
            :param osc_range: maximum number of oscillators
            :param iterate: shortcut to apply the method iterate() on the initialized DecomposedOscillatorModel
            :param **kwargs: key-word pair arguments for iterate()
        """
        if noise_start is None:
            noise_start = fs / 2 - 20
            if noise_start <= 0:
                noise_start = fs / 2 - 5
        assert noise_start > 0, 'Please redefine noise_start. Entered or calculated frequency is less than zero.'
        R = estimate_r(y, fs, freq_cutoff=noise_start)

        # populate instance attributes
        self.noise_start = noise_start
        self.osc_range = osc_range
        self.keep_param = ()

        self.added_osc = initialize_allosc(fs, y, R, osc_range=osc_range)
        self.fitted_osc = []
        self.ll = []
        self.ebic = []
        self.knee_index = None

        self.iterate(**kwargs) if iterate else None  # perform learning iterations

    def reset(self):
        """ Re-instantiate the instance object with the same data """
        self.__init__(self.added_osc[0].y, self.added_osc[0].Fs, noise_start=self.noise_start, osc_range=self.osc_range)

    def iterate(self, freq_res=None, keep_param=(), preiterate=False,
                R_hp=0.1, model_select='knee', plot_fit=False, verbose=None):
        """
        Decomposed Oscillator Search (dOsc) Algorithm

        Inputs:
        :param self: DecomposedOscillatorModel class instance
        :param freq_res: not used but kept for inheritance by the OscillatorModel class
        :param keep_param: a tuple of strings for parameters to keep and not update
        :param preiterate: prefit the oscillators to be added and sort in descending order of peak PSD
        :param R_hp: hyperparameter in Inverse Gamma Prior determining weight on prior mode compared to MLE
        :param model_select: 'knee', 'max', or 'ebic' to select with the knee, maximum likelihood, or minimum BIC
        :param plot_fit: plot the fitted oscillator spectra at each iteration
        :param verbose: not used but kept for inheritance by the OscillatorModel class
        """
        # Prefit the AR eigenmodes as oscillators and sort them in descending order of peak PSD
        self.added_osc = self.preiterate(keep_param=keep_param, R_hp=R_hp) if preiterate else self.added_osc

        while len(self.fitted_osc) < self.osc_range:
            o1 = self.added_osc[0].copy()
            _ = [o1.append(x) for x in self.added_osc[1:len(self.fitted_osc) + 1]]

            # Run EM iterations
            priors = o1.initialize_priors(kappa=0, R_hyperparameter=R_hp, R_sigma2=o1.R[0, 0], Q_sigma2='MLE')
            for _ in range(50):
                _ = o1.m_estimate(**o1.dejong_filt_smooth(EM=True), priors=priors, keep_param=keep_param)

            if plot_fit:
                fig = o1.visualize_freq('theoretical', bw=1)
                fig.get_axes()[0].set_title('Iteration %d' % len(self.fitted_osc))
                fig.set_tight_layout(True)

            self.fitted_osc.append(o1.copy(drop_y=True))
            self.ll.append(o1.dejong_filt_smooth(return_dict=True)['logL'].sum())
            self.ebic.append(ebic_calc(osc=o1, ll=self.ll[-1], osc_range=self.osc_range, gamma=0))

        # Find the knee_index for the selected model of fitted oscillators
        if model_select == 'knee':
            self.knee_index = get_knee(self.ll)
        elif model_select == 'max':
            self.knee_index = np.argmax(self.ll)
        elif model_select == 'ebic':
            self.knee_index = np.argmin(self.ebic)
        else:
            raise ValueError('model_select must be either "knee", "max", or "ebic"')

    def preiterate(self, keep_param=(), R_hp=0.1):
        """ Prefit the oscillators to be added and sort in descending order of peak PSD """
        # first fit with all oscillators to be added at once
        o1 = self.added_osc[0].copy()
        _ = [o1.append(x) for x in self.added_osc[1:]]  # use all available oscillators

        priors = o1.initialize_priors(kappa=0, R_hyperparameter=R_hp, R_sigma2=o1.R[0, 0], Q_sigma2='MLE')
        for _ in range(50):
            _ = o1.m_estimate(**o1.dejong_filt_smooth(EM=True), priors=priors, keep_param=keep_param)
        o1.fill_components()  # fill component parameters after learning

        # get the theoretical spectral peaks for each fitted oscillator
        peak_psd = np.squeeze(np.array([c._theoretical_spectrum(f_hz=c.freq[0])[0] for c in o1.components]))

        # create a new list of oscillators to be added in descending order of peak PSD
        add_indices = np.argsort(peak_psd)[::-1]
        added_osc = [o1.components[add_indices[0]].copy()]  # keep only one copy of y
        for ii in range(1, len(add_indices)):
            added_osc.append(o1.components[add_indices[ii]].copy(drop_y=True))

        return added_osc

    def get_knee_osc(self):
        """ Return the OscillatorModel at the selected knee index """
        return self.fitted_osc[self.knee_index]

    def get_residual(self):
        """ Compute the one-step prediction error from the best model """
        y = self.added_osc[0].y
        osc = self.get_knee_osc()
        residual, _ = get_innovations(osc, y=y)
        return residual

    def get_residual_fit(self, residual=None):
        """
        Fit a straight line to the residuals from the best model using
        a theoretical AR model spectrum
        """
        # Obtain the theoretical spectrum of a fitted AR model
        residual = self.get_residual() if residual is None else residual
        a, p = fit_ar(np.squeeze(residual), self.osc_range * 2 - 1)
        ar_psd, ar_hz = get_ar_psd(self.get_knee_osc().Fs, a, p)

        # Fit a straight line in semi-log space, i.e., log(PSD) against linear frequency (Hz)
        from scipy import stats
        slope, intercept, *_ = stats.linregress(ar_hz, 10 * np.log10(ar_psd))

        return slope, intercept

    def get_fit_spectra(self, bw=1):
        """ Compute the theoretical and empirical spectra of oscillators from the best model """
        selected_osc = self.get_knee_osc()
        spectra_theo = selected_osc.oscillator_spectra('theoretical')
        kalman_out = selected_osc.dejong_filt_smooth(y=self.added_osc[0].y, return_dict=True)
        spectra_empi = selected_osc.oscillator_spectra('actual', kalman_out['x_t_n'][:, 1:], bw)
        return spectra_theo, spectra_empi

    def diagnose_residual_acf(self):
        """
        Diagnose autocorrelations that might be present in the residuals
        using a series of Portmanteau tests (meaning they have unspecified
        alternative hypothesis but a well-defined null hypothesis)
        """
        from statsmodels.stats.stattools import durbin_watson
        from statsmodels.stats.diagnostic import acorr_ljungbox, acorr_lm
        from statsmodels.tsa.stattools import bds

        residual = self.get_residual()
        osc = self.get_knee_osc()

        # Durbin-Watson test for AR(1) process
        dw_dstat = durbin_watson(residual)
        # Ljung-Box test for autocorrelations up to 1s lags
        lb_df = acorr_ljungbox(residual, lags=[osc.Fs], model_df=3*osc.ncomp)
        # Lagrange Multiplier test for autocorrelations up to second order
        lm_results = acorr_lm(residual, nlags=2, ddof=3*osc.ncomp)
        # Brock-Dechert-Scheinkman test for iid. of residuals at second order embedding
        bds_results = bds(residual, max_dim=3)

        # Create a dataframe to store the results
        df = pd.DataFrame({'Durbin-Watson d': dw_dstat,
                           'Ljung-Box': lb_df['lb_pvalue'].values[0],
                           'Lagrange Multiplier': lm_results[1],
                           'Brock-Dechert-Scheinkman': bds_results[1][1]},
                          index=['p-value'])

        return df

    def diagnose_residual_norm(self):
        """ Diagnose whether residuals are white noise using normality tests """
        from scipy.stats import zscore, shapiro, normaltest, anderson, cramervonmises

        residual = self.get_residual()

        # Shapiro-Wilk test for normality
        sw_results = shapiro(residual)
        # D'Agostino and Pearson test for normality
        dp_results = normaltest(residual, nan_policy='omit')
        # Anderson-Darling test for normality using critical value at 5% significance level
        ad_results = anderson(residual, dist='norm')
        ad_reject_null = ad_results.statistic >= ad_results.critical_values[2]
        # Cramer-von Mises test for normality after z-scoring the residuals
        cvm_results = cramervonmises(zscore(residual), cdf='norm')

        # Create a dataframe to store the results
        df = pd.DataFrame({'Shapiro-Wilk': sw_results.pvalue,
                           'D\'Agostino and Pearson': dp_results.pvalue,
                           'Anderson-Darling H': ad_reject_null,
                           'Cramer-von Mises': cvm_results.pvalue},
                          index=['p-value'])

        return df

    def diagnose_residual(self):
        """ Generate diagnostic plots and statistics on the residuals """
        from IPython.display import display

        _ = self.plot_residual()
        _ = self.plot_residual_fit()
        _ = self.plot_acf()
        _ = self.plot_pacf()

        print('>>> Portmanteau tests for autocorrelations:')
        display(self.diagnose_residual_acf())

        print('>>> Normality tests for white residuals:')
        display(self.diagnose_residual_norm())

        return

    def plot_residual(self, **kwargs):
        """ Plot the best fit model and its residuals against white noises """
        fig = plot_residual(self.get_knee_osc(), self.added_osc[0].y, self.get_residual(), **kwargs)
        return fig

    def plot_residual_fit(self, ax=None):
        """ Plot the one-step prediction error fitted with an AR model and a straight line """
        residual = self.get_residual()
        slope, intercept = self.get_residual_fit(residual=residual)
        a, p = fit_ar(np.squeeze(residual), self.osc_range * 2 - 1)

        fig, _ = plot_innovations(self.get_knee_osc(), self.added_osc[0].y,
                                  residual, a, p, ax=ax, figsize=(6.4, 4.8), counter=self.knee_index)
        ax1 = fig.axes[1] if ax is None else ax[1]
        fig = plot_fit_line(fig, slope, intercept, ax=ax1)
        return fig

    def plot_acf(self, ax=None):
        """ Plot the autocorrelation function of the residuals """
        fig = plot_acf(self.get_residual(), Fs=self.get_knee_osc().Fs, ax=ax)
        return fig

    def plot_pacf(self, ax=None):
        """ Plot the partial autocorrelation function of the residuals """
        fig = plot_pacf(self.get_residual(), Fs=self.get_knee_osc().Fs, ax=ax)
        return fig

    def plot_fit_spectra(self, bw=1, ax=(None, None)):
        """ Plot the fitted theoretical and empirical spectra """
        selected_osc = self.get_knee_osc()
        if ax[0] is not False:
            fig_theo = selected_osc.visualize_freq('theoretical', bw=bw, y=self.added_osc[0].y, ax=ax[0])
        else:
            fig_theo = None
        if ax[1] is not False:
            fig_empi = selected_osc.visualize_freq('actual', bw=bw, y=self.added_osc[0].y, ax=ax[1])
        else:
            fig_empi = None
        return fig_theo, fig_empi

    def plot_log_likelihoods(self, ax=None):
        """ Plot the trajectory of log likelihoods from iterations """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()
        ax.plot(range(len(self.fitted_osc)), self.ll, '-*', label='__nolegend__')
        ax.scatter(self.knee_index, self.ll[self.knee_index], color='red', zorder=2, label='knee')
        ax.set_xlabel('Index of Model')
        ax.grid('on')
        ax.set_title('Log Likelihood')
        return fig

    def plot_mtm(self, **kwargs):
        """ Plot multitaper spectrogram and mean spectrum """
        return plot_mtm(self, **kwargs)

    def plot_trace(self, ax=None):
        """ Plot the time trace """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()
        ax.plot(self.added_osc[0].y[0, :])
        ax.set_xlabel('Samples')
        ax.set_ylabel('Voltage')
        ax.set_title('Time Trace')
        return fig

    def __repr__(self):
        """ Unambiguous and concise representation when calling DecomposedOscillatorModel() """
        return 'DecOsc(' + str(len(self)) + ')<' + hex(id(self))[-4:] + '>'

    def __str__(self):
        """ Helpful information when calling print(DecomposedOscillatorModel()) """
        print_str = "number of oscillators = %d\n " % len(self)
        if self.fitted_osc:
            print_str += str(self.fitted_osc[-1])
        else:
            print_str += 'There is no fitted oscillator. Initial oscillator:\n '
            print_str += str(self.added_osc[0])
        return print_str

    def __len__(self):
        return self.added_osc[0].ncomp + len(self.fitted_osc) - 1



"""
Author: Mingjian He <mh1@stanford.edu>

Testing functions for exact inference signal processing methods in somata/exact_inferences/dp_func.py
"""

import numpy as np
from codetiming import Timer
from test_load_data import _load_data  # type: ignore


def test_forward_backward(show_plot=False):
    from somata.exact_inference import forward_backward

    input_data = _load_data("forward_backward_test_data.mat")

    p_ht_v1_t, p_ht_v1_T, p_ht_tmin1_v1_T, logL = forward_backward(A=input_data['A'], py_x=input_data['qt_m'],
                                                                   p1=input_data['p1'], compute_edge=True)

    # Verify that the results are close within tolerance
    assert np.allclose(p_ht_v1_t.flatten(), input_data['filt'].flatten()), 'Filtering results are off.'
    assert np.allclose(p_ht_v1_T.flatten(), input_data['smooth'].flatten()), 'Smoothing results are off.'
    assert np.allclose(p_ht_tmin1_v1_T.flatten(), input_data['edges'].flatten()), 'Edge marginal results are off.'
    assert np.allclose(logL.flatten(), input_data['logL'].flatten()), 'Log likelihood results are off.'

    if show_plot:
        import matplotlib.pyplot as plt

        # Compare the differences in filtered, smoothed, edge marginals, and log likelihood results
        fig, axs = plt.subplots(2, 2, tight_layout=True)

        # filtered
        var_diff = p_ht_v1_t.flatten() - input_data['filt'].flatten()
        axs[0, 0].hist(var_diff)
        axs[0, 0].set_title('Filtered', fontweight="bold")

        # smoothed
        var_diff = p_ht_v1_T.flatten() - input_data['smooth'].flatten()
        axs[0, 1].hist(var_diff)
        axs[0, 1].set_title('Smoothed', fontweight="bold")

        # edge marginals
        var_diff = p_ht_tmin1_v1_T.flatten() - input_data['edges'].flatten()
        axs[1, 0].hist(var_diff)
        axs[1, 0].set_title('Edge Marginals', fontweight="bold")

        # log likelihood
        var_diff = logL.flatten() - input_data['logL'].flatten()
        axs[1, 1].hist(var_diff)
        axs[1, 1].set_title('logL', fontweight="bold")


def test_viterbi():
    from somata.exact_inference import viterbi

    input_data = _load_data("viterbi_test_data.mat")

    viterbi_path, bin_path = viterbi(A=input_data['A'], py_x=input_data['py_x'], p1=input_data['p1'])

    # Verify that the results are identical
    assert (viterbi_path + 1 == input_data['viterbi_path']).all(), 'Viterbi path results are off.'
    assert (bin_path == input_data['bin_path']).all(), 'Binary path results are off.'


def test_kalman(show_plot=False):
    from somata.exact_inference import kalman

    input_data = _load_data("kalman_inputs.mat")

    # Run the python kalman function on the given inputs
    with Timer():
        x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, x_t_tmin1, P_t_tmin1, fy_t_interp = kalman(**input_data)

    # Compare with MATLAB results
    output_data = _load_data("kalman_outputs.mat")

    # Verify that the results are close within tolerance
    for elem in output_data.keys():
        py_var = locals()[elem]
        ml_var = output_data[elem]
        if elem == 'fy_t_interp':
            assert np.isnan(py_var), "NaN fy_t_interp failed in Python."
            assert np.isnan(ml_var), "Nan fy_t_interp failed in MATLAB."
        else:
            assert np.allclose(py_var.flatten(), ml_var.flatten()), 'Results of kalman differ from MATLAB outputs.'

    if show_plot:
        import matplotlib.pyplot as plt

        fig, axs = plt.subplots(5, 2, tight_layout=True)
        fig.suptitle('kalman', y=0.995, fontweight="bold", fontsize=16)
        ax_num = 0

        for elem in output_data.keys():
            py_var = locals()[elem]
            ml_var = output_data[elem]

            axs_index = np.unravel_index(ax_num, axs.shape, 'F')

            if elem == 'fy_t_interp':
                assert np.isnan(py_var), "NaN fy_t_interp failed in Python."
                assert np.isnan(ml_var), "Nan fy_t_interp failed in MATLAB."
                axs[axs_index].axvline(x=0, color='b', linestyle='--', linewidth=4)
                axs[axs_index].set_xlim([-1, 1])
            else:
                assert py_var.shape == ml_var.shape, "Dimensions don't match between MATLAB and Python outputs"
                if (py_var == ml_var).all():
                    axs[axs_index].axvline(x=0, color='b', linestyle='--', linewidth=4)
                    axs[axs_index].set_xlim([-1, 1])
                else:
                    var_diff = py_var.flatten() - ml_var.flatten()
                    axs[axs_index].hist(var_diff)

            axs[axs_index].set_title(elem, fontweight="bold")
            ax_num += 1


def test_djkalman(show_plot=False):
    from somata.exact_inference import djkalman

    input_data = _load_data("kalman_inputs.mat")

    # Run the python djkalman function on the given inputs
    with Timer():
        x_t_n, P_t_n, P_t_tmin1_n, logL, x_t_t, P_t_t, K_t, \
            x_t_tmin1, P_t_tmin1, fy_t_interp = djkalman(**input_data, skip_interp=False)

    # Compare with MATLAB results
    output_data = _load_data("djkalman_outputs.mat")

    # Verify that the results are close within tolerance
    for elem in output_data.keys():
        assert np.allclose(locals()[elem].flatten(), output_data[elem].flatten()), \
            'Results of djkalman differ from MATLAB outputs.'

    if show_plot:
        import matplotlib.pyplot as plt

        fig, axs = plt.subplots(5, 2, tight_layout=True)
        fig.suptitle('djkalman', y=0.995, fontweight="bold", fontsize=16)
        ax_num = 0

        for elem in output_data.keys():
            py_var = locals()[elem]
            ml_var = output_data[elem]

            axs_index = np.unravel_index(ax_num, axs.shape, 'F')

            assert py_var.shape == ml_var.shape, "Dimensions don't match between MATLAB and Python outputs."
            if (py_var == ml_var).all():
                axs[axs_index].axvline(x=0, color='b', linestyle='--', linewidth=4)
                axs[axs_index].set_xlim([-1, 1])
            else:
                var_diff = py_var.flatten() - ml_var.flatten()
                axs[axs_index].hist(var_diff)

            axs[axs_index].set_title(elem, fontweight="bold")
            ax_num += 1


def test_inverse(dim=128):
    from somata.exact_inference import inverse

    np.random.seed(1)

    # Test with a small symmetric positive definite matrix
    A = np.random.rand(dim, dim)
    A = A @ A.T + np.eye(dim)
    A_np_inv = np.linalg.inv(A)

    assert np.allclose(A_np_inv.flatten(), inverse(A, approach='gaussian').flatten()), 'Gaussian inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse(A, approach='cholesky').flatten()), 'Cholesky inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse(A, approach='qr').flatten()), 'QR inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse(A, approach='svd').flatten()), 'SVD inversion is off.'

    # Test with a diagonal matrix
    A = np.eye(dim) * np.random.rand(dim)
    A_np_inv = np.linalg.inv(A)

    assert np.allclose(A_np_inv.flatten(), inverse(A, approach='gaussian').flatten()), 'Gaussian inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse(A, approach='cholesky').flatten()), 'Cholesky inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse(A, approach='qr').flatten()), 'QR inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse(A, approach='svd').flatten()), 'SVD inversion is off.'

    # Test with a singular diagonal matrix
    A = np.eye(dim) * np.random.rand(dim)
    A_man_inv = 1.0 / np.diagonal(A)
    A[-1, -1] = 0.0
    A_man_inv[-1] = 0.0
    A_man_inv = np.eye(dim) * A_man_inv

    assert np.allclose(A_man_inv.flatten(), inverse(A, approach='svd').flatten()), 'SVD inversion is off.'
    assert A.dtype == inverse(A, approach='svd').dtype, 'SVD inversion changed the dtype.'


def test_inverse_torch(dim=128, atol=1e-3):
    from somata.exact_inference import inverse_torch
    import torch

    np.random.seed(1)

    # Test with a symmetric positive definite matrix
    A = np.random.rand(dim, dim)
    A = A @ A.T + np.eye(dim)
    A_np_inv = np.linalg.inv(A)
    A = torch.as_tensor(data=A, dtype=torch.float32)

    assert np.allclose(A_np_inv.flatten(), inverse_torch(A, approach='gaussian').cpu().numpy().flatten(),
                       atol=atol), 'Gaussian inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse_torch(A, approach='cholesky').cpu().numpy().flatten(),
                       atol=atol), 'Cholesky inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse_torch(A, approach='qr').cpu().numpy().flatten(),
                       atol=atol), 'QR inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse_torch(A, approach='svd').cpu().numpy().flatten(),
                       atol=atol), 'SVD inversion is off.'

    # Test with a diagonal matrix
    A = np.eye(dim) * np.random.rand(dim)
    A_np_inv = np.linalg.inv(A)
    A = torch.as_tensor(data=A, dtype=torch.float32)

    assert np.allclose(A_np_inv.flatten(), inverse_torch(A, approach='gaussian').cpu().numpy().flatten(),
                       atol=atol), 'Gaussian inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse_torch(A, approach='cholesky').cpu().numpy().flatten(),
                       atol=atol), 'Cholesky inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse_torch(A, approach='qr').cpu().numpy().flatten(),
                       atol=atol), 'QR inversion is off.'
    assert np.allclose(A_np_inv.flatten(), inverse_torch(A, approach='svd').cpu().numpy().flatten(),
                       atol=atol), 'SVD inversion is off.'

    # Test with a singular diagonal matrix
    A = np.eye(dim) * np.random.rand(dim)
    A_man_inv = 1.0 / np.diagonal(A)
    A[-1, -1] = 0.0
    A_man_inv[-1] = 0.0
    A_man_inv = np.eye(dim) * A_man_inv
    A = torch.as_tensor(data=A, dtype=torch.float32)

    assert np.allclose(A_man_inv.flatten(), inverse_torch(A, approach='svd').cpu().numpy().flatten(),
                       atol=atol), 'SVD inversion is off.'
    assert A.dtype == inverse_torch(A, approach='svd').dtype, 'SVD inversion changed the dtype.'


def test_logdet():
    from somata.exact_inference import logdet

    np.random.seed(1)

    for ii in range(8):
        dim = 2 ** ii

        # Test with a symmetric positive semi-definite matrix
        A = np.random.rand(dim, dim)
        A = A @ A.T
        assert np.allclose(np.log(np.linalg.det(A)), logdet(A)
                           ), 'Log determinant results are off for positive semi-definite matrices.'

        # Test with a diagonal matrix
        A = np.eye(dim) * np.random.rand(dim)
        assert np.allclose(np.log(np.linalg.det(A)), logdet(A)
                           ), 'Log determinant results are off for diagonal matrices.'


def test_logdet_torch(rtol=1e-3, atol=1e-3):
    from somata.exact_inference import logdet_torch
    import torch

    np.random.seed(1)

    for ii in range(8):
        dim = 2 ** ii

        # Test with a symmetric positive semi-definite matrix
        A = np.random.rand(dim, dim)
        A = A @ A.T
        numpy_logdet = np.log(np.linalg.det(A))
        A = torch.as_tensor(data=A, dtype=torch.float32)
        assert np.allclose(numpy_logdet, logdet_torch(A).cpu().numpy(),
                           rtol=rtol, atol=atol), 'Log determinant results are off for positive semi-definite matrices.'

        # Test with a diagonal matrix
        A = np.eye(dim) * np.random.rand(dim)
        numpy_logdet = np.log(np.linalg.det(A))
        A = torch.as_tensor(data=A, dtype=torch.float32)
        assert np.allclose(numpy_logdet, logdet_torch(A).cpu().numpy(),
                           rtol=rtol, atol=atol), 'Log determinant results are off for diagonal matrices.'


if __name__ == "__main__":
    test_forward_backward()
    test_viterbi()
    test_kalman()
    test_djkalman()
    test_inverse()
    test_inverse_torch()
    test_logdet()
    test_logdet_torch()
    print('DP function tests finished without exception.')



"""
Author: Mingjian He <mh1@stanford.edu>

dmap_em module contains dynamic source localization codes used in SOMATA
"""

import torch
import mne
import numpy as np
import numbers
from collections.abc import Iterable
from copy import deepcopy
from joblib import cpu_count
from somata.basic_models import StateSpaceModel
from somata.exact_inference import run_em, inverse
from somata.utils import estimate_r
from somata.source_loc.source_loc_utils import smart_djkalman_conv_torch, get_whitener
from scipy.linalg import qr, block_diag, sqrtm
from scipy.stats import norm
import matplotlib.pyplot as plt

# Turn off tensor 32bit precision to allow full 32bit floating point
torch.backends.cuda.matmul.allow_tf32 = False
torch.backends.cudnn.allow_tf32 = False


class SourceLocModel(object):
    """
    SourceLocModel is an object class dedicated to dynamic source
    localization inference in SOMATA, supporting all somata classes
    and multiple components
    """
    nstate = None  # hidden state dimension
    ncomp = None  # number of components
    nchannel = None  # number of channels
    nsource = None  # number of sources
    components = None  # a list of independent components
    comp_nstates = None  # a list of state dimensions of components
    proj = None  # projection matrix on observed data
    lfm = None  # lead field gain matrix
    G = None  # observation matrix
    Ps = None  # permutation matrix
    Ms = None  # mixing matrix
    B = None  # adjacency matrix
    Q_basis = None  # orthonormal basis of adjacency matrix
    ssm_params = None  # state-space model parameters
    em_log = None  # dictionary of EM iteration numbers and changes in log likelihoods

    def __init__(self, components, proj=None, lfm=None, B=None, src=None, fwd=None, **kwargs):
        """
        Constructor method for SourceLocModel class
        :param components: a list of independent components in the model
        :param proj: projection matrix on observed data (default to average referencing)
        :param lfm: forward model lead field gain matrix (lfm < fwd)
        :param B: adjacency matrix for the basis of source noise covariance (B < src)
        :param src: MNE SourceSpaces instance with anatomical information (src < fwd)
        :param fwd: MNE Forward instance with source information (fwd takes the highest priority)
        """
        # Process the list of independent components underlying all sources
        self.ncomp = len(components)
        self.components = list(components) if isinstance(components, Iterable) else [components]
        self.comp_nstates = [x.default_G.shape[1] for x in self.components]

        # Process Forward if provided (takes precedence over lfm and src inputs)
        if fwd is not None:
            assert lfm is None and src is None, 'Forward provided in fwd. lfm and src inputs will be overwritten.'
            lfm, src = self._process_fwd_input(fwd)

        # Set up forward model and related attributes
        self.lfm = deepcopy(np.asanyarray(lfm, dtype=np.float64))
        self.nchannel, self.nsource = lfm.shape
        self.nstate = sum(self.comp_nstates) * self.nsource
        self.G = self._build_observation_matrix()
        self.Ps = [self._build_permutation_matrix(nstate=x) for x in self.comp_nstates]

        # Process the projection matrix
        if proj is None:  # default to average referencing projector
            indi_vec = np.ones((self.nchannel, 1), dtype=np.float64)
            self.proj = np.eye(self.nchannel, dtype=np.float64) - indi_vec @ indi_vec.T / (indi_vec.T @ indi_vec)
        else:
            self.proj = proj

        # Process SourceSpaces if provided to set up spatial dependencies
        if src is not None:
            assert B is None, 'SourceSpaces provided in src. adjacency matrix input B will be overwritten.'
            assert len(src[0]['vertno']) + len(src[1]['vertno']) == self.nsource, \
                'Total vertno is different from nsource.'
            d1, d2, m1, m2, scale, approach, orthonormal = self._process_src_scalars(**kwargs)
            B = self._build_adjacency_matrix(src=src, d1=d1, d2=d2, scale=scale, approach=approach)
            self.Ms = [self._build_mixing_matrix(nstate=x, src=src, m1=m1, m2=m2) for x in self.comp_nstates]
        else:
            self.Ms = [np.eye(x * self.nsource, dtype=np.float64) for x in self.comp_nstates]
            orthonormal = None

        # Create the basis for noise covariance across sources
        self.B = np.eye(self.nsource, dtype=np.float64) if B is None else B
        self.Q_basis = self._get_q_basis(self.B, orthonormal=orthonormal)

    # Dunder methods - magic methods
    def __repr__(self):
        """ Unambiguous and concise representation when calling SourceLocModel() """
        return 'Src(' + str(self.ncomp) + ')<' + hex(id(self))[-4:] + '>'

    def __str__(self):
        """ Helpful information when calling print(SourceLocModel()) """
        print_str = "<Src object at " + hex(id(self)) + ">\n " + \
                    "{0:8} = {1: <5} ".format("nstate", str(self.nstate)) + \
                    "{0:8} = {1}\n ".format("ncomp", str(self.ncomp)) + \
                    "{0:8} = {1: <5} ".format("nchannel", str(self.nchannel)) + \
                    "{0:8} = {1}\n ".format("nsource", str(self.nsource)) + \
                    "{0:12} = {1}\n ".format("comp_nstates", str(self.comp_nstates)) + \
                    "components = {}\n ".format(str(self.components)) + \
                    "{0:3}.shape = {1}\n ".format("lfm", [str(x.shape) if x is not None else 'None' for x
                                                          in [self.lfm]][0]) + \
                    "{0:3}.shape = {1}\n ".format("G", [str(x.shape) if x is not None else 'None' for x
                                                        in [self.G]][0]) + \
                    "{0:8} = {1} ".format("ssm_params", [str(x.keys()) if x is not None else 'None' for x
                                                         in [self.ssm_params]][0])
        return print_str

    def __len__(self):
        return self.ncomp

    # EM methods - dynamic source localization with Expectation-Maximization algorithm
    def learn(self, y, rank=None, SNR=3, R=None, Q=None, mu0=None, S0=None,
              priors=None, Q_prior_option='MLE', R_prior_option='MLE',
              dtype=torch.float32, update_param=('F', 'Q', 'R'), keep_param=(),
              max_iter=10, logL_thresh=1e-6, show_pbar=True):
        """
        Dynamic Source Localization with MAP EM Learning

        The algorithm is based on the following reference, with substantial
        modifications made to flexibly handle all somata classes and compute
        more efficiently and robustly for other SSM structures beyond AR1.

        Reference:
            Pirondini, E., Babadi, B., Obregon-Henao, G., Lamus, C., Malik, W. Q.,
            Hämäläinen, M. S., & Purdon, P. L. (2017). Computationally efficient
            algorithms for sparse, dynamic solutions to the EEG source localization
            problem. IEEE Transactions on Biomedical Engineering, 65(6), 1359-1372.

        Inputs:
        :param y: observed data (row major, should be multivariate)
        :param rank: rank of observed data (use of mne.compute_rank() is recommended)
        :param SNR: signal-to-noise ratio of amplitude in scalp measurement
        :param R: observation noise covariance matrix or a scalar
        :param Q: state noise covariance matrix
        :param mu0: initial state mean vector
        :param S0: initial state covariance matrix
        :param priors: a list of dictionaries specifying priors for each component, if None -> initialize_priors()
        :param Q_prior_option: a string selecting a prior to use for MAP estimating Q
        :param R_prior_option: a string selecting a prior to use for MAP estimating R
        :param dtype: numerical precision data type for GPU processing
        :param update_param: a tuple of strings for parameters to update
        :param keep_param: a tuple of strings for parameters to keep and not update in M step
        :param max_iter: maximal number of EM iterations in dMAP-EM
        :param logL_thresh: threshold below which EM iteration stops
        :param show_pbar: show progress bar during EM iterations
        """
        # Initialize state-space model parameters
        R, rank = self._initialize_r(y=y, rank=rank, R=R, cutoff_frequency=35)
        mu0 = np.zeros((self.nstate, 1), dtype=np.float64) if mu0 is None else mu0
        S0, sigma2_S0 = self._initialize_S0(R=R, rank=rank, S0=S0, SNR=SNR, diagonal=True)
        Q, sigma2_Q = self._initialize_q(S0=S0, sigma2_S0=sigma2_S0, Q=Q)

        # Move attribute matrices to cuda device
        self.Ps = [torch.as_tensor(P, dtype=dtype).cuda() for P in self.Ps]
        self.Ms = [torch.as_tensor(M, dtype=dtype).cuda() for M in self.Ms]
        if isinstance(self.Q_basis, list):
            self.Q_basis = [torch.as_tensor(q, dtype=dtype).cuda() for q in self.Q_basis]
        else:
            self.Q_basis = torch.as_tensor(self.Q_basis, dtype=dtype).cuda()

        # Initialize E-step using a dictionary of state-space model parameters
        y = torch.as_tensor(y, dtype=dtype).cuda()
        F = self._build_transition_matrix()
        ssm_params = {'F': F, 'Q': Q, 'mu0': mu0, 'S0': S0, 'G': self.G, 'R': R}
        for elem in ssm_params.keys():  # convert to torch data type for GPU processing
            ssm_params[elem] = torch.as_tensor(ssm_params[elem], dtype=dtype).cuda()
        self.ssm_params = ssm_params

        # Initialize priors dictionary for M-step
        priors = self.initialize_priors(sigma2_Q=sigma2_Q, sigma2_S0=sigma2_S0, Q_prior_option=Q_prior_option,
                                        R=ssm_params['R'].detach().clone(), T=y.shape[1],
                                        R_prior_option=R_prior_option) if priors is None else priors

        # Learn with EM algorithm
        e_kwargs = {'logL_list': [float('-inf')], 'rank': rank}
        m_kwargs = {'priors': priors, 'update_param': update_param, 'keep_param': keep_param}
        self.em_log = run_em(self, y=y, e_kwargs=e_kwargs, m_kwargs=m_kwargs, max_iter=max_iter,
                             stop_thresh=logL_thresh, ignore_numerr=True, return_dict=True, show_pbar=show_pbar)

        # Final Kalman smoothing to get hidden state estimates
        x_t_n, P_t_n, *_, = smart_djkalman_conv_torch(y=y, rank=rank, proj=self.proj, **self.ssm_params)

        return x_t_n.cpu().numpy(), P_t_n.cpu().numpy()

    def e_step(self, y, rank=None, logL_list=None):
        """ E-step in dMAP-EM, exposed for run_em() """
        x_t_n, P_t_n, P_t_tmin1_n, logL, break_conv, *_ = smart_djkalman_conv_torch(
            y=y, rank=rank, proj=self.proj, **self.ssm_params)
        e_logL = torch.sum(logL)
        stop_var = abs(e_logL - logL_list[-1])  # due to convergent Kalman, may not monotonically increase
        logL_list.append(e_logL)

        # Announce the flag of reaching numerical precision limit
        stop_var = True if break_conv or torch.isnan(e_logL) else stop_var

        return {'x_t_n': x_t_n, 'P_t_n': P_t_n, 'P_t_tmin1_n': P_t_tmin1_n}, stop_var

    def m_step(self, y, x_t_n=None, P_t_n=None, P_t_tmin1_n=None, priors=None,
               update_param=('F', 'Q', 'R'), keep_param=()):
        """ M-step in dMAP-EM, exposed for run_em() """
        # Obtain boolean flags of scopes of updates
        update_FQ, update_mu0S0 = self._m_step_scope(update_param, keep_param)

        # Initialize parameters
        T = y.shape[1]
        if priors is None:
            priors = [None] * self.ncomp
        elif isinstance(priors, dict):
            priors = [priors]  # so that it can be indexed to position 0

        # Attempt to skip sums of squares computation in the steady-state version
        if update_FQ:
            A = P_t_n * T + x_t_n[:, :-1] @ x_t_n[:, :-1].T
            B = P_t_tmin1_n * T + x_t_n[:, 1:] @ x_t_n[:, :-1].T
            C = P_t_n * T + x_t_n[:, 1:] @ x_t_n[:, 1:].T
        else:
            A, B, C = (None, None, None)

        # Update parameters for each independent component -- F, Q, mu0, S0 (component specific priors)
        if update_FQ or update_mu0S0:
            # Initialize variables to store updated parameters
            if 'Q' in update_param and 'Q' not in keep_param:
                Q_new = torch.zeros(P_t_n.shape, dtype=P_t_n.dtype).cuda()
            else:
                Q_new = None

            if 'mu0' in update_param and 'mu0' not in keep_param:
                mu0_new = torch.zeros((self.nstate, 1), dtype=x_t_n.dtype).cuda()
            elif 'S0' in update_param and 'S0' not in keep_param:
                mu0_new = self.ssm_params['mu0']
            else:
                mu0_new = None

            if 'S0' in update_param and 'S0' not in keep_param:
                S0_new = torch.zeros(P_t_n.shape, dtype=P_t_n.dtype).cuda()
            else:
                S0_new = None

            # Iterate through independent components
            for ii in range(self.ncomp):
                current_component = self.components[ii]
                start_idx = sum(self.comp_nstates[:ii]) * self.nsource
                end_idx = sum(self.comp_nstates[:ii + 1]) * self.nsource
                if update_FQ:
                    A_tmp = A[start_idx:end_idx, start_idx:end_idx]
                    B_tmp = B[start_idx:end_idx, start_idx:end_idx]
                    C_tmp = C[start_idx:end_idx, start_idx:end_idx]
                else:
                    A_tmp, B_tmp, C_tmp = (None, None, None)

                # Call the _m_update_<param> methods specific to the component subclass
                if 'F' in update_param and 'F' not in keep_param:
                    # Combine sums of squares across sources for the same component
                    ABC = self._combine_src_ss(self.comp_nstates[ii], self.Ms[ii], A_tmp, B_tmp, C_tmp)
                    # noinspection PyProtectedMember
                    current_component.F = current_component._m_update_f(A=ABC['A'], B=ABC['B'], priors=priors[ii])
                    current_component.update_comp_param()  # update component specific parameters

                if 'Q' in update_param and 'Q' not in keep_param:
                    # Compute the sum of squares inside the trace term of complete loglikelihood that involves Q
                    F_tmp = self._build_transition_matrix(components=[current_component], Ms=[self.Ms[ii]])
                    Q_ss = C_tmp - B_tmp @ F_tmp.T - F_tmp @ B_tmp.T + F_tmp @ A_tmp @ F_tmp.T

                    # Re-order sum of squares into Q_basis block diagonal form
                    P = self.Ps[ii]
                    Q_ss = P.T @ Q_ss @ P

                    # Call the _m_update_q_src method specific to the component subclass
                    # noinspection PyProtectedMember
                    Q_new_ii = current_component._m_update_q_src(Q_ss=Q_ss, T=T, Q_basis=self.Q_basis,
                                                                 nsource=self.nsource,
                                                                 nstate=self.comp_nstates[ii],
                                                                 priors=priors[ii])

                    # Re-order Q_new_ii back into the Kalman form
                    Q_new[start_idx:end_idx, start_idx:end_idx] = P @ Q_new_ii @ P.T

                if 'mu0' in update_param and 'mu0' not in keep_param:
                    # noinspection PyProtectedMember
                    mu0_new[start_idx:end_idx, 0] = \
                        current_component._m_update_mu0_src(x_0_n=x_t_n[start_idx:end_idx, 0],
                                                            nstate=self.comp_nstates[ii])[:, 0]

                if 'S0' in update_param and 'S0' not in keep_param:
                    x_0_n = x_t_n[start_idx:end_idx, 0][:, None]
                    mu0_tmp = mu0_new[start_idx:end_idx, 0][:, None]

                    # Compute the posterior covariance at t=0
                    S0_ss = P_t_n + x_0_n @ x_0_n.T - x_0_n @ mu0_tmp.T - mu0_tmp @ x_0_n.T + mu0_tmp @ mu0_tmp.T

                    # Re-order the covariance into Q_basis block diagonal form
                    P = self.Ps[ii]
                    S0_ss = P.T @ S0_ss @ P

                    # Re-use the _m_update_q_src method specific to the component subclass
                    # noinspection PyProtectedMember
                    S0_new_ii = current_component._m_update_q_src(Q_ss=S0_ss, T=1, Q_basis=self.Q_basis,
                                                                  nsource=self.nsource,
                                                                  nstate=self.comp_nstates[ii],
                                                                  priors=None)

                    # Re-order S0_new_ii back into the Kalman form
                    S0_new[start_idx:end_idx, start_idx:end_idx] = P @ S0_new_ii @ P.T

        # Update the dictionary of ssm parameters
        if 'F' in update_param and 'F' not in keep_param:
            F_new = self._build_transition_matrix()
            self.ssm_params.update({'F': F_new})

        if 'Q' in update_param and 'Q' not in keep_param:
            # noinspection PyUnboundLocalVariable
            self.ssm_params.update({'Q': Q_new})

        if 'mu0' in update_param and 'mu0' not in keep_param:
            # noinspection PyUnboundLocalVariable
            self.ssm_params.update({'mu0': mu0_new})

        if 'S0' in update_param and 'S0' not in keep_param:
            # noinspection PyUnboundLocalVariable
            self.ssm_params.update({'S0': S0_new})

        if 'R' in update_param and 'R' not in keep_param:
            R_new = self._m_update_r(y=y, x_t_n=x_t_n, P_t_n=P_t_n, T=T, priors=priors[0])
            self.ssm_params.update({'R': R_new})

        return

    def _combine_src_ss(self, nstate, M, A_tmp, B_tmp, C_tmp):
        """ Combine sums of squares across sources for the same component """
        # Apply the mixing matrix associated with transition matrix
        A_hat = M @ A_tmp @ M.T
        B_hat = B_tmp @ M.T
        C_hat = C_tmp

        # Combine sums of squares across sources
        A = torch.zeros((nstate, nstate), dtype=A_tmp.dtype).cuda()
        B = torch.zeros((nstate, nstate), dtype=B_tmp.dtype).cuda()
        C = torch.zeros((nstate, nstate), dtype=C_tmp.dtype).cuda()
        for vidx in range(self.nsource):
            A += A_hat[vidx * nstate:(vidx + 1) * nstate, vidx * nstate:(vidx + 1) * nstate]
            B += B_hat[vidx * nstate:(vidx + 1) * nstate, vidx * nstate:(vidx + 1) * nstate]
            C += C_hat[vidx * nstate:(vidx + 1) * nstate, vidx * nstate:(vidx + 1) * nstate]
        return {'A': A, 'B': B, 'C': C}

    def _m_update_r(self, y=None, x_t_n=None, P_t_n=None, T=None, priors=None):
        """ Update observation noise covariance """
        G = self.ssm_params['G']
        pred_err = y - G @ x_t_n[:, 1:]
        R_ss = pred_err @ pred_err.T + G @ (P_t_n * T) @ G.T

        # Due to convergent Kalman, R_ss diagonal could be negative. We lower-bound it to 0
        lower_bound_index = torch.where(torch.lt(R_ss, 0))[0]
        R_ss[lower_bound_index, lower_bound_index] = 0.
        # if len(lower_bound_index) > 0:
        #     print('Encountered negative diagonal elements when updating R!')

        # noinspection PyProtectedMember
        if StateSpaceModel._m_update_if_mle('R_wishart', priors):
            # MLE
            R_new = R_ss / T
        else:
            # MAP with Wishart prior
            R_new = (R_ss + priors['R_wishart']['v'] * priors['R_wishart']['R_prior']) / (
                    T + priors['R_wishart']['v'] + priors['R_wishart']['p'] + 1)

        return R_new

    # Matrix building methods - construct important matrices for dmap_em()
    def _build_observation_matrix(self):
        """ Construct observation matrix G """
        stacked_G = []
        for comp in self.components:
            stacked_G.append(np.hstack([self.lfm[:, ii][:, None] * comp.default_G
                                        for ii in range(self.lfm.shape[1])]))
        G = np.hstack(stacked_G)
        assert G.shape[0] == self.nchannel, 'Incorrect channel dimension of observation matrix G.'
        assert G.shape[1] == self.nstate, 'Incorrect state dimension of observation matrix G.'
        return G

    def _build_transition_matrix(self, components=None, Ms=None):
        """ Construct transition matrix F """
        components = self.components if components is None else components
        Ms = self.Ms if Ms is None else Ms
        F_blocks = [torch.block_diag(*[torch.as_tensor(components[x].F, dtype=Ms[x].dtype).cuda()] * self.nsource
                                     ) @ Ms[x] for x in range(len(components))]
        F = torch.block_diag(*F_blocks).cuda()
        return F

    def _build_permutation_matrix(self, nstate):
        """
        Construct the elementary row/column permutation matrix
        to transform a block diagonal Q basis into the Q block
        for one independent component.

        The relation is: P @ Q_basis @ P.T = Q
        """
        P = np.zeros((self.nsource * nstate, self.nsource * nstate), dtype=int)
        for vidx in range(self.nsource):
            for j in range(nstate):
                to_idx = vidx * nstate + j
                from_idx = j * self.nsource + vidx
                P[to_idx, from_idx] = 1
        return P

    def _build_mixing_matrix(self, nstate, src, m1=0.5, m2=0.5):
        """
        Construct the mixing matrix to linearly combine hidden states
        from first- and second-order neighbors when evolving the state
        equation, i.e., provide a right-multiplying mixing transform
        to the transition matrix.

        If a diagonal transition matrix is of the form:
             X_t = F @ X_t-1
        With a mixing matrix M, the state equation becomes:
             X_t = F @ M @ X_t-1

        The weighting on a source and its neighbors is as following:
            m1 * x + m2 * (d_N1 * N1(x)) + (1 - m1 - m2) * (d_N2 * N2(x))

        N.B.: Here the d_N1 are normalized to sum to 1 among first-order
        neighbors. Similarly, d_N2 are normalized to sum to 1 among
        second-order neighbors. For a justification of this, see Appendix A
        in the following reference.

        Reference:
            Lamus, C., Hämäläinen, M. S., Temereanca, S., Brown, E. N., & Purdon,
            P. L. (2012). A spatiotemporal dynamic distributed solution to the
            MEG inverse problem. NeuroImage, 63(2), 894-909.
        """
        # Default weighting ignores second-order neighbors
        m1 = 1.0 if m1 is None else m1
        m2 = 0.0 if m2 is None else m2
        assert m1 + m2 <= 1, 'The total weights must be less than or equal to 1.'

        # Compute the pairwise-distance between all sources in use
        if src[0]['dist'] is None:
            mne.add_source_space_distances(src, n_jobs=cpu_count())

        # Create the index mapping between vertex and source numbering
        vert_to_source, source_to_vert = self._vertex_source_mapping(src)

        # Define neighbors of all used vertices in original vertex numbering
        neighbors = self._define_neighbors(src)

        # Build the mixing matrix
        M = np.zeros((nstate * self.nsource, nstate * self.nsource), dtype=np.float64)

        for hemi in range(len(src)):
            for vidx in source_to_vert[hemi]:
                vert = source_to_vert[hemi][vidx]  # vertex indexing

                # Add the mixing block for the source itself
                M[vidx * nstate:(vidx + 1) * nstate, vidx * nstate:(vidx + 1) * nstate] = \
                    m1 * np.eye(nstate, dtype=np.float64)

                for order, m in zip(['first', 'second'], (m2, 1 - m1 - m2)):
                    vert_neighbor = np.asarray([vert_to_source[hemi].get(x, float('nan'))
                                                for x in neighbors[hemi][vert][order]])
                    vert_distance = np.asarray([src[hemi]['dist'][vert, x]
                                                for x in neighbors[hemi][vert][order]])

                    # Filter out neighbor vertices that are not sources
                    valid_idx = np.invert(np.isnan(vert_neighbor))
                    vert_neighbor = vert_neighbor[valid_idx].astype(dtype=int)
                    vert_distance = vert_distance[valid_idx]

                    # Normalize the inverse distances so they sum to one
                    inv_distance = 1 / vert_distance
                    inv_distance = inv_distance / inv_distance.sum()

                    # Add the mixing block for the neighbors
                    for v, d in zip(vert_neighbor, inv_distance):
                        M[vidx * nstate:(vidx + 1) * nstate, v * nstate:(v + 1) * nstate] = \
                            m * d * np.eye(nstate, dtype=np.float64)

        return M

    def _build_adjacency_matrix(self, src, d1=0.5, d2=0.25, scale=None, plot_kernel=False, approach=None):
        """
        Create adjacency matrix to capture spatial dependencies
        between sources based on the distance and neighboring
        relationship, i.e., the local topology of SourceSpaces
        """
        # Default scaling on first- and second-order neighbors
        d1 = 0.0 if d1 is None else d1
        d2 = 0.0 if d2 is None else d2
        approach = approach or 'normalize_neighbors'

        # Compute the pairwise-distance between all sources in use
        if src[0]['dist'] is None:
            mne.add_source_space_distances(src, n_jobs=cpu_count())

        # Create the index mapping between vertex and source numbering
        vert_to_source, source_to_vert = self._vertex_source_mapping(src)

        # Define neighbors of all used vertices in original vertex numbering
        neighbors = self._define_neighbors(src)

        # Build the adjacency matrix
        """
        There are different ways to construct the adjacency matrix and obtain
        a set of diagonal heavy kernels for update state noise covariance
        matrix Q.
        """
        if approach == 'normalize_neighbors':
            # (1) normalized and scaled distances from neighbors (Pirondini et al. 2017)
            B = np.eye(self.nsource, dtype=np.float64)

            # Populate the adjacency matrix with normalized inverse distances
            for hemi in range(len(src)):
                for vidx in source_to_vert[hemi]:
                    vert = source_to_vert[hemi][vidx]  # vertex indexing

                    for order, delta in zip(['first', 'second'], (d1, d2)):
                        vert_neighbor = np.asarray([vert_to_source[hemi].get(x, float('nan'))
                                                    for x in neighbors[hemi][vert][order]])
                        vert_distance = np.asarray([src[hemi]['dist'][vert, x]
                                                    for x in neighbors[hemi][vert][order]])

                        # Filter out neighbor vertices that are not sources
                        valid_idx = np.invert(np.isnan(vert_neighbor))
                        vert_neighbor = vert_neighbor[valid_idx].astype(dtype=int)
                        vert_distance = vert_distance[valid_idx]

                        # Normalize the inverse distances so they sum to one
                        inv_distance = 1 / vert_distance
                        inv_distance = inv_distance / inv_distance.sum()

                        # Fill the scaled and normalized inverse distances
                        B[vidx, vert_neighbor] = delta * inv_distance

        elif approach == 'symmetric_neighbors':
            # (2) symmetric and diagonal scaling from neighbors (Behtash Babadi)
            B = np.zeros((self.nsource, self.nsource), dtype=np.float64)
            N = np.zeros((self.nsource, self.nsource), dtype=int)

            # STEP 1: Populate the adjacency matrix with inverse distances
            for hemi in range(len(src)):
                for vidx in source_to_vert[hemi]:
                    vert = source_to_vert[hemi][vidx]  # vertex indexing

                    for order, n in zip(['first', 'second'], (1, 2)):
                        vert_neighbor = np.asarray([vert_to_source[hemi].get(x, float('nan'))
                                                    for x in neighbors[hemi][vert][order]])
                        vert_distance = np.asarray([src[hemi]['dist'][vert, x]
                                                    for x in neighbors[hemi][vert][order]])

                        # Filter out neighbor vertices that are not sources
                        valid_idx = np.invert(np.isnan(vert_neighbor))
                        vert_neighbor = vert_neighbor[valid_idx].astype(dtype=int)
                        vert_distance = vert_distance[valid_idx]

                        # Store the inverse distances and neighborhood status
                        B[vidx, vert_neighbor] = 1 / vert_distance
                        N[vidx, vert_neighbor] = n  # mark the neighbors for scaling

            # STEP 2: Add the sum of each row / column as the diagonal
            B_diagonal = B.sum(axis=0)
            B[range(self.nsource), range(self.nsource)] = B_diagonal

            # STEP 3: Normalize by the square root of diagonal row- and column-wise
            sqrt_B_diag = np.sqrt(B_diagonal)
            B = B / sqrt_B_diag[:, None] / sqrt_B_diag[None, :]

            # STEP 4: Scale the first- and second-order neighbors
            B[N == 1] *= d1
            B[N == 2] *= d2

            # STEP 5: Sanity checks
            assert np.allclose(B, B.T), 'Adjacency matrix is not symmetric.'
            assert np.allclose(np.diag(B), np.ones(self.nsource)), 'Diagonal of adjacency matrix is not unity.'

        elif approach == 'symmetric_gaussian':
            # (3) symmetric and Gaussian kernels from all sources (Mingjian He and Proloy Das)
            B = np.zeros((self.nsource, self.nsource), dtype=np.float64)

            if scale is None:
                # Depending on the icosahedron resolution, set Gaussian std so that
                # the adjacency matrix entry for a "typically" distanced first-order
                # neighbor of a source is fixed at 0.25
                if src[0]['nuse'] < 1000:  # ico3 has 642 srcs per hemi
                    scale = np.sqrt(0.0124 ** 2 / (-2 * np.log(0.25)))  # std = 0.00745
                elif src[0]['nuse'] < 3000:  # ico4 has 2562 srcs per hemi
                    scale = np.sqrt(0.0062 ** 2 / (-2 * np.log(0.25)))  # std = 0.00372
                elif src[0]['nuse'] < 11000:  # ico5 has 10242 srcs per hemi
                    scale = np.sqrt(0.0031 ** 2 / (-2 * np.log(0.25)))  # std = 0.00186
                else:
                    raise ValueError('Ico resolution is non-standard. Cannot set the std of Gaussian kernel.')

            # Visualize the Gaussian kernel on top of src-src distances
            if plot_kernel:
                all_dist = np.hstack([src[0]['dist'].data, src[1]['dist'].data])
                fig, ax = plt.subplots()
                ax.hist(all_dist, bins=1000)
                plt.xlabel('Distance (m)', fontsize='x-large')
                plt.ylabel('Counts', fontsize='x-large')
                plt.title('Histogram of src-src distances', fontsize='xx-large')
                ax2 = ax.twinx()
                x = np.linspace(start=0, stop=all_dist.max(initial=None), num=10000)
                ax2.plot(x, norm.pdf(x, loc=0, scale=scale) / norm.pdf(0, loc=0, scale=scale), color='red')
                plt.ylabel('Gaussian kernel weights', fontsize='x-large')

            # Populate the adjacency matrix with normalized Gaussian kernels
            for hemi in range(len(src)):
                hemi_verts = src[hemi]['vertno']  # vertex indexing
                hemi_vidxs = [vert_to_source[hemi].get(x, float('nan')) for x in hemi_verts]
                for vidx, vert in zip(hemi_vidxs, hemi_verts):
                    B[vidx, hemi_vidxs] = norm.pdf(src[hemi]['dist'][vert, hemi_verts].toarray(),
                                                   loc=0, scale=scale) / norm.pdf(0, loc=0, scale=scale)

            # Sanity checks
            assert np.allclose(B, B.T), 'Adjacency matrix is not symmetric.'
            assert np.allclose(np.diag(B), np.ones(self.nsource)), 'Diagonal of adjacency matrix is not unity.'

        else:
            raise ValueError('Specified approach to build adjacency matrix is not recognized.')

        return B

    # Initialization methods - initialize parameters for dMAP-EM
    def initialize_priors(self, sigma2_Q=None, sigma2_S0=None, Q_prior_option='MLE',
                          R=None, T=None, R_prior_option='MLE'):
        """
        Initialize priors for covariance matrices
        used for dMAP-EM source localization
        """
        priors = [{} for _ in self.components]

        # Add source localization specific priors to each component
        for prior in priors:
            # State noise covariance
            if Q_prior_option == 'inverse gamma':
                """ Non-informative inverse gamma prior (Lamus et al. 2012) """
                E_theta = sigma2_Q  # magnitude of source activity
                Var_theta = 1 / E_theta  # Variance >> Expectation
                alpha = E_theta ** 2 / Var_theta + 2  # approximately 2
                beta = E_theta * (E_theta ** 2 / Var_theta + 1)  # approximately E_theta
                prior['theta_ig'] = {'alpha': alpha, 'beta': beta}
            elif Q_prior_option == 'log-sum':
                """ Sparse L0 log-sum prior (Pirondini et al. 2017) """
                prior['theta_l0'] = {'gamma': 1 / sigma2_S0}
            elif Q_prior_option == 'exponential':
                """ Sparse L1 exponential prior (Pirondini et al. 2017) """
                gamma = np.asarray([-np.log(1 - (x + 0.5) / self.nsource) for x in range(self.nsource)]
                                   ).sum() / (self.nsource * sigma2_S0)
                prior['theta_l1'] = {'gamma': gamma}
            elif Q_prior_option == 'jeffreys':
                """ Sparse Jeffreys prior (Pirondini et al. 2017) """
                prior['theta_jeff'] = True

            # Observation noise covariance - Wishart prior
            if R_prior_option == 'wishart':
                prior['R_wishart'] = {'R_prior': R, 'p': self.nchannel, 'v': T}

        return priors

    def _initialize_r(self, y, rank=None, R=None, cutoff_frequency=35, plot_freqz=False):
        """ Initialize observation noise covariance matrix R """
        assert y.shape[0] == self.nchannel, 'Dimensions of y are incompatible with lfm.'

        # Check the rank of observed data
        rank = self.nchannel if rank is None else rank
        assert rank <= self.nchannel, 'Incorrect rank number. Cannot be greater than the number of channels.'

        # Estimate the observation noise covariance matrix R
        if isinstance(R, numbers.Number):
            R = R * np.eye(self.nchannel, dtype=np.float64)

        elif R is None:
            R = estimate_r(y, self.components[0].Fs, freq_cutoff=cutoff_frequency, plot_freqz=plot_freqz)

        # Update rank to be the rank of projected data
        eig, eigvec = mne.utils.linalg.eigh(R, overwrite_a=False)
        eig = np.ones(eig.shape, dtype=np.float64)
        eig[:-rank] = 0.0
        temp_R = self.proj @ eigvec * eig @ eigvec.T @ self.proj.T
        eig, _ = mne.utils.linalg.eigh(temp_R, overwrite_a=True)
        rank = np.sum(eig > 1e-2)

        return R, rank

    def _initialize_S0(self, R, rank=None, S0=None, SNR=3, diagonal=True):
        """
        Initialize state covariance matrix S0 at t=0

        Inputs:
        :param R: observation noise covariance matrix
        :param rank: rank of projected data
        :param SNR: amplitude SNR in scalp measurement
        :param diagonal: whether to initialize S0 as a diagonal matrix

        About SNR:
            - MNE uses amplitude SNR of 3, i.e., power SNR of 9
            - Lamus et al. 2012 uses power SNR of 5
            - Pirondini et al. 2017 uses power SNR of 3
        """
        if S0 is not None:
            sigma2_S0 = np.trace(S0) / S0.shape[0]

        else:
            # Lamus et al. 2012
            W, _ = get_whitener(R, rank=rank, proj=self.proj)
            WG = W @ self.G
            scale_S0 = rank / np.sum(WG ** 2)

            """
            The above is a more efficient and accurate calculation of the expression:

                scale_S0 = self.nchannel / np.trace(self.G @ self.G.T @ np.linalg.inv(R))

            MNE-Python as of v0.24.1 used the following expression:

                scale_S0 = np.trace(R) / np.trace(self.G @ self.G.T)

            that is the same as above when data is whitened.

            Note that Pirondini et al. 2017 used the second expression
            without whitening the data, resulting in an approximation
            with only the diagonal of R, which is inaccurate.
            """

            lambda2 = 1 / SNR ** 2
            sigma2_S0 = scale_S0 / lambda2

            S0_blocks = []
            for ii in range(self.ncomp):
                current_component = self.components[ii]
                if diagonal:
                    E = np.eye(self.nsource, dtype=np.float64)
                else:
                    E = self.Q_basis @ self.Q_basis.T
                S0_tmp = sigma2_S0 * current_component.get_default_q(components=current_component, E=E)
                S0_blocks.append(self.Ps[ii] @ S0_tmp @ self.Ps[ii].T)  # re-order into the Kalman form

            S0 = block_diag(*S0_blocks)

        return S0, sigma2_S0

    @staticmethod
    def _initialize_q(S0, sigma2_S0, Q=None, src_scalp_SNR_ratio=0.1):
        """ Initialize state covariance matrix Q """
        if Q is not None:
            sigma2_Q = np.trace(Q) / Q.shape[0]

        else:
            # Lamus et al. 2012 introduced this heuristic scaling, assuming
            # SNR in the source space is 1/10 of the SNR in scalp measurement
            Q = src_scalp_SNR_ratio * S0
            sigma2_Q = src_scalp_SNR_ratio * sigma2_S0

        return Q, sigma2_Q

    @staticmethod
    def _process_fwd_input(fwd):
        """ Process MNE Forward input to constructor """
        # Ensure fixed normal direction is used for lead field gain matrix
        if fwd['sol']['data'].shape == fwd['_orig_sol'].shape:
            fwd = mne.convert_forward_solution(fwd, force_fixed=True, surf_ori=True)

        lfm = fwd['sol']['data']
        src = fwd['src']

        return lfm, src

    @staticmethod
    def _process_src_scalars(**kwargs):
        """
        Process the scalars used to construct adjacency
        matrix for state noise covariance matrix and to
        construct mixing matrix for transition matrix
        """
        d1 = kwargs['d1'] if 'd1' in kwargs else None
        d2 = kwargs['d2'] if 'd2' in kwargs else None
        m1 = kwargs['m1'] if 'm1' in kwargs else None
        m2 = kwargs['m2'] if 'm2' in kwargs else None
        scale = kwargs['scale'] if 'scale' in kwargs else None
        approach = kwargs['approach'] if 'approach' in kwargs else None
        orthonormal = kwargs['orthonormal'] if 'orthonormal' in kwargs else None
        return d1, d2, m1, m2, scale, approach, orthonormal

    @staticmethod
    def _vertex_source_mapping(src):
        """
        Create dictionaries to map between vertex numbering contained
        as 'vertno' attribute of MNE SourceSpaces and numbering of
        sources in the forward model lead field gain matrix, i.e.,
        the numbering of columns in lfm
        """
        assert len(src) == 2, 'Expecting two hemispheres contained in SourceSpaces instance.'
        vert_to_source, source_to_vert = ([{}] * len(src), [{}] * len(src))

        # Left (hemi=0) and right (hemi=1) hemispheres
        for hemi in range(len(src)):
            nsource = src[hemi]['nuse']
            source_indices = (np.asarray(range(nsource)) + src[0]['nuse']).tolist() if hemi == 1 else range(nsource)
            vertex_indices = src[hemi]['vertno']
            vert_to_source[hemi] = dict(zip(vertex_indices, source_indices))
            source_to_vert[hemi] = dict(zip(source_indices, vertex_indices))

        return vert_to_source, source_to_vert

    @staticmethod
    def _define_neighbors(src):
        """
        Define the indices of first and second order neighbors of all
        vertices in the source triangulation as in the 'use_tris'
        attribute of MNE SourceSpaces
        """
        assert len(src) == 2, 'Expecting two hemispheres contained in SourceSpaces instance.'
        neighbors = [{}] * len(src)

        # Left (hemi=0) and right (hemi=1) hemispheres
        for hemi in range(len(src)):
            hemi_neighbors: dict = {}

            # Define first-order neighbors
            for tri in src[hemi]['use_tris']:
                for vert in tri:
                    # Set default key-value pairs if not already present
                    hemi_neighbors.setdefault(vert, {})
                    hemi_neighbors[vert].setdefault('first', [])

                    # Add the other two vertices in the triangulation to first-order neighbors
                    first_list = hemi_neighbors[vert]['first']
                    vert_neighbors = tri.copy().tolist()
                    vert_neighbors.remove(vert)
                    hemi_neighbors[vert]['first'] = np.unique(first_list + vert_neighbors).tolist()

            # Define second-order neighbors
            for vert in hemi_neighbors:
                first_list = hemi_neighbors[vert]['first']
                unique_second = np.unique(np.hstack([hemi_neighbors[x]['first'] for x in first_list]))
                hemi_neighbors[vert]['second'] = np.setdiff1d(unique_second, first_list + [vert]).tolist()

            neighbors[hemi] = hemi_neighbors

        return neighbors

    @staticmethod
    def _get_q_basis(B, orthonormal=None):
        """
        Decompose the adjacency matrix into an orthonormal
        basis for noise covariance, or use the matrix square
        root of the adjacency matrix as a non-orthonormal kernel
        """
        orthonormal = orthonormal or True
        if orthonormal:
            QR = qr(np.float64(B), pivoting=False)
            return QR[0]  # orthonormal basis, N.B. not eigen-basis!
        else:
            B_sqrt = sqrtm(B)
            return [B_sqrt, inverse(B_sqrt)]  # non-orthonormal kernel

    @staticmethod
    def update_theta(Q_ss=None, T=None, Q_basis=None, nsource=None, npart=None, priors=None):
        """ Update the expansion coefficients for the state noise covariance matrix Q """
        Theta = torch.zeros(nsource, dtype=Q_ss.dtype).cuda()
        for vidx in range(nsource):
            theta_ss = Q_basis[:, vidx] @ Q_ss @ Q_basis[:, vidx]

            # noinspection PyProtectedMember
            if StateSpaceModel._m_update_if_mle(['theta_ig', 'theta_l0', 'theta_l1', 'theta_jeff'], priors):
                # MLE
                Theta[vidx] = theta_ss / (npart * T)
            elif 'theta_ig' in priors:
                # inverse gamma prior
                alpha = priors['theta_ig']['alpha']
                beta = priors['theta_ig']['beta']
                Theta[vidx] = (beta + (theta_ss / npart) / 2) / (alpha + T / 2 + 1)
            elif 'theta_l0' in priors:
                gamma = priors['theta_l0']['gamma']
                d = theta_ss * gamma / npart - T
                Theta[vidx] = (d / 2 + torch.sqrt(
                    d ** 2 / 4 + 2 * gamma * (2 / (npart ** 2) + (T / npart) / 2) * theta_ss)
                               ) / (gamma * (4 / npart + T))
            elif 'theta_l1' in priors:
                # exponential prior
                gamma = priors['theta_l1']['gamma']
                Theta[vidx] = (torch.sqrt(T ** 2 + 8 * theta_ss * gamma / (npart ** 2)) - T) / (4 * gamma / npart)
            elif 'theta_jeff' in priors:
                # Jeffreys prior
                Theta[vidx] = theta_ss / (npart * T + 2)
            else:
                raise RuntimeError('Could not set m_update rule for theta. Please check.')

        # assert np.all(Theta > 0), 'Theta coefficients should be non-negative.'
        # Due to convergent Kalman, Theta could be negative. We lower-bound it to 0
        Theta[Theta < 0] = 0

        return Theta

    @staticmethod
    def _m_step_scope(update_param, keep_param):
        """ Sort out the scopes of updates in m_step() """
        # noinspection PyProtectedMember
        update_FQ, update_mu0S0 = StateSpaceModel._m_estimate_scope(update_param, keep_param)
        assert 'G' not in update_param, 'Observation matrix should not be updated during source localization.'
        return update_FQ, update_mu0S0



"""
Author: Mingjian He <mh1@stanford.edu>

vb module contains variational bayesian methods for switching inference
"""

from somata.basic_models import StateSpaceModel as Ssm
from somata.exact_inference import forward_backward, viterbi, run_em, logdet, inverse
from somata.switching import switching
import numpy as np
from codetiming import Timer


class VBSwitchModel(object):
    """
    VBSwitchModel is an object class created to help
    Variational Bayesian Switching Models inference
    """
    ssm_array = None
    K = None
    T = None
    A = None
    h_t_m = None
    q_t_m = None
    logL_bound = None
    x_t_n_all = None
    P_t_n_all = None

    def __init__(self, ssm_array, y=None):
        """
        Constructor method for VBSwitchModel class
        :param ssm_array: an array of Ssm class objects or a single object with stacked models
        :param y: observed data
        """
        self.ssm_array, self.K, self.T = Ssm.setup_array(ssm_array, y=y)

    def __repr__(self):
        """ Unambiguous and concise representation when calling VBSwitchModel() """
        return 'Vbs(' + str(len(self.ssm_array)) + ')<' + hex(id(self))[-4:] + '>'

    def __str__(self):
        """ Helpful information when calling print(VBSwitchModel()) """
        print_str = "Vbs model " + str(self.ssm_array)
        return print_str

    def learn(self, y=None, dwell_prob=0.99, A=None, keep_param=(),
              maxE_iter=100, maxVB_iter=100, h_t_thresh=1e-6, q_t_thresh=1e-6,
              shared_R=False, shared_comp=False, priors_all='auto', normalize_q_t=False,
              plot_E=False, verbose=True, return_dict=None, original=False, show_pbar=True):
        """
        Variational Bayesian learning of segmental state-space
        models with Markov switching

        This is an implementation of Ghahramani & Hinton 2000
        learning algorithm with variational bayesian approximation.
        This algorithm has a generalized EM structure where the E
        step will lower-bound the posterior distribution of the
        hidden variables (both continuous hidden state and discrete
        switching variables), and the M step will re-estimate the
        state-space model (linear dynamical system) and hidden Markov
        model parameters using Shumway & Stoffer 1982 EM on the SSM
        and Baum-Welch algorithm on the HMM.

        Figure 5: Learning algorithm for switching state-space models
        [E step] - Repeat until convergence of KL(Q||P):
                E.1 Compute q_t_m from the prediction error of
                    state-space model m on observation Y_t
                E.2 Compute h_t_m using the forward-backward algorithm
                    on the HMM, with observation probabilities q_t_m
                E.3 For m = 1 to M (we use K instead of M)
                    Run Kalman smoothing recursions, using the data
                    weighted by h_t_m

        [M step]
                M.1 Re-estimate parameters for each state-space model
                    using the data weighted by h_t_m
                M.2 Re-estimate parameters for the switching process
                    using Baum-Welch update equations.

        Reference:
            He, M., Das, P., Hotan, G., & Purdon, P. L. (2023). Switching
            state-space modeling of neural signal dynamics. PLoS
            Computational Biology, 19(8), e1011395.

            Ghahramani, Z., & Hinton, G. E. (2000). Variational learning
            for switching state-space models. Neural computation, 12(4),
            831-864.

        If original is False, we use a new implementation of the original
        G&H2000 learning algorithm with variational bayesian approximation.
        This new implementation introduces various improvements across
        the VB learning iterations. These features include:
            - smart initialization of E-step with interpolated density
              and avoid deterministic annealing
            - semi-hard (converged h_t), soft (smoothed estimates of
              discrete HMM using interpolated density), and hard
              (Viterbi) segmentations are available
            - parallelized Kalman filtering and smoothing in De Jong
              version that avoids inverting conditional state noise
              covariance P_t_tmin1 and handles degenerate matrices
              automatically
            - parameters for shared components can be jointly estimated
            - MAP estimation instead of ML to incorporate priors

        Inputs:
        :param y: observed data
        :param dwell_prob: dwell probability for HMM transitions to initialize A
        :param A: HMM transition matrix, if provided will override dwell_prob
        :param keep_param: a tuple of strings for parameters to keep and not update
        :param maxE_iter: maximal number of iterations to run for fixed-point iteration in E step
        :param maxVB_iter: maximal number of EM iterations in VB learning
        :param h_t_thresh: threshold below which EM iteration stops
        :param q_t_thresh: threshold below which E step fixed-point iteration stops
        :param shared_R: boolean flag to do joint estimation of observation noise
        :param shared_comp: False to do independent updates, or a tuple of lists of indices for which component
                            in each model to estimate jointly. E.g. ([1,1], [2,3]) will force the 1st component
                            of model 1 to be linked to the 1st component of model 2, and the 2nd component of
                            model 1 to be linked to the 3rd component of model 2. Lists inside the tuple must
                            have the same length as the number of models in ssm_array
        :param priors_all: a list of lists of dictionaries specifying priors for each component,
                           if 'auto' -> Auto MAP, if None -> MLE
        :param normalize_q_t: whether to normalize interpolated density at the first E-step iteration
        :param plot_E: boolean flag to plot during fixed-point iteration in E step
        :param verbose: boolean flag for printing messages of changes in h_t_m and time taken
        :param return_dict: None -> no return, True -> return dict, False -> return tuple of variables
        :param original: boolean flag to use the original Ghahramani & Hinton algorithm
        :param show_pbar: show progress bar during EM iterations

        Outputs:
        :returns: a dictionary or tuple of variables depending on return_dict
            - h_t_m: a 2D array of model responsibilities for each model at each time point
            - h_t_m_soft: a 2D array of soft segmentations for each model at each time point
            - h_t_m_hard: a 2D array of hard segmentations for each model at each time point
            - q_t_m: a 2D array of emission probability densities for each model at each time point
            - A: HMM transition matrix
            - ssm_array: an array of Ssm class objects
            - VB_iter: number of EM iterations
            - logL_bound: a list of lower bounds on log likelihood
            - x_t_n_all: a list of smoothed estimates (posterior) of state mean
            - P_t_n_all: a list of smoothed estimates (posterior) of state conditional covariance
        """
        t = Timer(name="time taken", text="{name}: {seconds:.2f} seconds")
        t.start()

        """ Initialization """
        self.A = abs(np.eye(self.K, dtype=np.float64) - np.ones((self.K, self.K), dtype=np.float64)
                     * (1-dwell_prob) / (self.K-1)) if A is None else A
        self.logL_bound = []
        if original:
            shared_R = True
            shared_comp = False
            priors_all = None
            self.h_t_m = np.ones((self.K, self.T), dtype=np.float64) / self.K  # initialize with equal responsibilities
        else:
            self.h_t_m = float('inf')
        # Initialize priors for SSM if 'auto' is specified
        priors_all = [x.initialize_priors() for x in self.ssm_array] if priors_all == 'auto' else priors_all

        e_kwargs = {'original': original, 'maxE_iter': maxE_iter, 'q_t_thresh': q_t_thresh,
                    'normalize_q_t': normalize_q_t, 'plot_E': plot_E}
        m_kwargs = {'priors_all': priors_all, 'keep_param': keep_param,
                    'shared_R': shared_R, 'shared_comp': shared_comp}

        """ EM iterations """
        VB_iter, _ = run_em(self, y=y, e_kwargs=e_kwargs, m_kwargs=m_kwargs,
                            max_iter=maxVB_iter, stop_thresh=h_t_thresh, ignore_numerr=True, show_pbar=show_pbar)

        """ Switching linear segmentation """
        # Provide different segmentation methods (soft, semi-hard (VB h_t_m), hard)
        h_t_m_soft, fy_t = switching(self.ssm_array, method='ab pari', A=self.A)  # soft segmentation
        _, h_t_m_hard = viterbi(self.A, fy_t, ignore_numerr=True)  # apply Viterbi on parallel interpolated density

        """ EOF """
        if verbose:
            print('VBSwitchModel.learn()')
            t.stop()
            print('iterations:', str(VB_iter))

        if return_dict is None:
            pass
        elif return_dict:
            return {'h_t_m': self.h_t_m, 'h_t_m_soft': h_t_m_soft, 'h_t_m_hard': h_t_m_hard, 'q_t_m': self.q_t_m,
                    'A': self.A, 'ssm_array': self.ssm_array, 'VB_iter': VB_iter, 'logL_bound': self.logL_bound,
                    'x_t_n_all': self.x_t_n_all, 'P_t_n_all': self.P_t_n_all}
        else:
            return self.h_t_m, h_t_m_soft, h_t_m_hard, self.q_t_m, \
                self.A, self.ssm_array, VB_iter, self.logL_bound, self.x_t_n_all, self.P_t_n_all

    def e_step(self, y=None, original=False, maxE_iter=100, q_t_thresh=1e-6, normalize_q_t=False, plot_E=False):
        """
        Generalized E step in VB learning, exposed for run_em()

        Depending on whether the original algorithm is used, the E step
        has a different initialization of the fixed-point iteration
        """
        # Save the current h_t_m to update EM stopping variable at the end
        last_h_t_m = self.h_t_m

        # Initialize the fixed-point iteration
        delta_q_t_m = float('inf')
        self.q_t_m = float('inf')
        E_iter = 0
        E_logL_bound = []

        # Different initialization based on whether the original algorithm is used
        if original:
            method = 'kalman'
            temperature = 100  # initial temperature parameter for deterministic annealing
            # Initialize E step using converged h_t_m from the last EM iteration as a warm start
            x_t_n_all, P_t_n_all, *_ = Ssm.par_kalman(self.ssm_array, y=y, method=method, R_weights=1/self.h_t_m)
        else:
            method = 'dejong'
            temperature = 1  # no deterministic annealing

        # minimize KL(Q||P)
        while delta_q_t_m > q_t_thresh and E_iter < maxE_iter:
            # Store q_t_m to check for convergence of E step
            E_iter += 1
            last_q_t_m = self.q_t_m

            # E.1 Compute q_t_m using smoothed estimates
            if E_iter == 1 and not original:
                # Initialize q_t_m in E.1 using interpolated density
                _, P_t_n_all, _, _, _, _, _, _, _, fy_t_interp_all = Ssm.par_kalman(self.ssm_array, y=y,
                                                                                    method=method, skip_interp=False)
                self.q_t_m = np.vstack(fy_t_interp_all)

                # Normalize interpolated density to better initialize q_t_m
                if normalize_q_t:
                    # Scale q_t_m such that the sum of interpolated densities are equal across models
                    sum_density_weights = self.q_t_m.sum(axis=1)
                    self.q_t_m = self.q_t_m * (min(sum_density_weights) / sum_density_weights)[:, None]

                    # Scale q_t_m by the effective smoothed conditional covariance
                    P_t_n_weights = np.zeros(self.K, dtype=np.float64)
                    for m in range(self.K):
                        P_t_n_weights[m] = np.squeeze(
                            self.ssm_array[m].G @ P_t_n_all[m][:, :, self.T//2] @ self.ssm_array[m].G.T)
                    self.q_t_m = self.q_t_m * (min(P_t_n_weights) / P_t_n_weights)[:, None]

                g_t_m = np.log(self.q_t_m)
            else:
                # noinspection PyUnboundLocalVariable
                self.q_t_m, g_t_m = _compute_q_t_m(self.ssm_array, x_t_n_all, P_t_n_all, temperature=temperature)

            # E.2 Compute h_t_m using forward-backward algorithm
            _, self.h_t_m, h_t_tmin1_m, logL_HMM = forward_backward(A=self.A, py_x=self.q_t_m, compute_edge=True)

            # Modify h_t_m and h_t_tmin1_m with deterministic annealing
            self.h_t_m = self.h_t_m / temperature
            h_t_tmin1_m = h_t_tmin1_m / temperature
            temperature = temperature / 2 + 0.5  # decreasing towards asymptote at temperature=1

            # E.3 Run Kalman smoothing recursions with weighted data
            x_t_n_all, P_t_n_all, P_t_tmin1_n_all, logL_all, *_ = \
                Ssm.par_kalman(self.ssm_array, y=y, method=method, R_weights=1/self.h_t_m)
            logL_SSM = np.vstack(logL_all)

            # Check for convergence using the average change in elements of q_t_m
            delta_q_t_m = np.mean(abs(last_q_t_m - self.q_t_m))

            # Compute the negative variational free energy / lower bound on likelihood / evidence lower bound (ELBO)
            E_logL_bound.append(_compute_logl_bound(self.ssm_array, logL_HMM, logL_SSM, self.h_t_m, g_t_m))

            # Plotting E step for the first two models
            _ = _plot_e(self.h_t_m, self.q_t_m, E_iter) if plot_E else None

        # Finish the E step
        self.logL_bound.append(E_logL_bound)
        # noinspection PyUnboundLocalVariable
        self.x_t_n_all = x_t_n_all
        # noinspection PyUnboundLocalVariable
        self.P_t_n_all = P_t_n_all

        # noinspection PyUnboundLocalVariable
        e_results = {'x_t_n_all': x_t_n_all, 'P_t_n_all': P_t_n_all, 'P_t_tmin1_n_all': P_t_tmin1_n_all,
                     'h_t_tmin1_m': h_t_tmin1_m}
        stop_var = np.mean(abs(last_h_t_m - self.h_t_m))

        return e_results, stop_var

    def m_step(self, y=None, x_t_n_all=None, P_t_n_all=None, P_t_tmin1_n_all=None, h_t_tmin1_m=None,
               priors_all=None, keep_param=(), shared_R=False, shared_comp=False):
        """ Generalized M step in VB learning, exposed for run_em() """
        # M.1 ML estimates of SSM parameters with weighted data
        _m1(self.ssm_array, x_t_n_all, P_t_n_all, P_t_tmin1_n_all, self.h_t_m, y=y,
            priors_all=priors_all, keep_param=keep_param, shared_R=shared_R, shared_comp=shared_comp)

        # M.2 ML estimate of transition matrix of HMM
        self.A, _ = _m2(self.h_t_m, h_t_tmin1_m)


def _compute_q_t_m(ssm_array, x_t_n_all, P_t_n_all, temperature=1, y=None):
    """
    Compute the un-normalized Gaussian density function
    of expected error for VB learning fixed-point iteration

    Reference:
        Ghahramani, Z., & Hinton, G. E. (2000). Variational learning
        for switching state-space models. Neural computation, 12(4),
        831-864.
    """
    # Array dimensions
    y = ssm_array[0].y if y is None else y
    K = len(ssm_array)  # number of models
    q, T = y.shape  # number of channels and number of time points
    approach = 'svd' if q >= 5 else 'gaussian'
    g_t_m = np.zeros((K, T), dtype=np.float64)

    for m in range(K):
        x_t_n = x_t_n_all[m][:, 1:]  # (index 1 corresponds to t=0, etc.)
        P_t_n = P_t_n_all[m][:, :, 1:]
        for ii in range(T):  # t=1 -> t=T
            g_t_m[m, ii] = 1/temperature * (
                    -0.5 * y[:, ii].T @ inverse(ssm_array[m].R, approach=approach) @ y[:, ii]
                    + y[:, ii].T @ inverse(ssm_array[m].R, approach=approach) @ ssm_array[m].G @ x_t_n[:, ii]
                    + -0.5 * np.trace(ssm_array[m].G.T @ inverse(ssm_array[m].R, approach=approach) @ ssm_array[m].G
                                      @ (P_t_n[:, :, ii] + x_t_n[:, ii][:, None] @ x_t_n[:, ii][:, None].T)))

    q_t_m = np.exp(g_t_m)  # Equation 4.13
    return q_t_m, g_t_m


def _compute_logl_bound(ssm_array, logL_HMM, logL_SSM, h_t_m, g_t_m):
    """
    Compute the bound on the marginal log likelihood of observed data.
    Also known as Evidence Lower BOund (ELBO) or negative variational
    free energy.
    """
    # Precompute the log determinant of R
    R_logdet = np.asarray([logdet(2 * np.pi * x.R) for x in ssm_array])

    # First term: determinant of R
    F1 = -0.5 * R_logdet @ h_t_m.sum(axis=1)

    # Second term: marginal log likelihood from HMM
    F2 = logL_HMM.sum()

    # Third term: -ht*gt
    F3 = -(h_t_m * g_t_m).sum()

    # Fourth term: marginal log likelihood from SSM combined with log(ht) to deal with infinity values
    F4_tmp = logL_SSM + 0.5 * (R_logdet[None, :].T - ssm_array[0].nchannel * np.log(h_t_m))
    F4_tmp[np.isinf(F4_tmp)] = float('nan')
    F4 = np.nansum(F4_tmp)

    logL_bound = F1 + F2 + F3 + F4
    return logL_bound


def _m1(ssm_array, x_t_n_all, P_t_n_all, P_t_tmin1_n_all, h_t_m, y=None, priors_all=None,
        keep_param=(), shared_R=False, shared_comp=False):
    """
    ML/MAP estimation of SSM parameters for the M.1 step.
    This updates SSM parameters with weighted data, and
    executes the M.1 step in VB learning of switching
    linear segments.

    Inputs:
    :param ssm_array: an array of Ssm class objects
    :param y: observed data
    :param x_t_n_all: a list of smoothed estimates (posterior) of state mean
    :param P_t_n_all: a list of smoothed estimates (posterior) of state conditional covariance
    :param P_t_tmin1_n_all: a list of smoothed estimates (posterior) of state lag1 conditional cross-covariance
    :param h_t_m: a 2D array of responsibility vectors
    :param priors_all: a list of lists of dictionaries specifying priors for each component, if None -> MLE
    :param keep_param: a tuple of strings for parameters to keep and not update
    :param shared_R: boolean flag to do joint estimation of observation noise
    :param shared_comp: False to do independent updates, or a tuple of lists of indices for which component
                        in each model to estimate jointly. E.g. ([1,1], [2,3]) will force the 1st component
                        of model 1 to be linked to the 1st component of model 2, and the 2nd component of
                        model 1 to be linked to the 3rd component of model 2. Lists inside the tuple must
                        have the same length as the number of models in ssm_array
    """
    y = ssm_array[0].y if y is None else y
    K = len(ssm_array)  # number of models
    T = y.shape[1]  # number of time points
    priors_all = [[None] * x.ncomp for x in ssm_array] if priors_all is None else priors_all

    """ Estimate model parameters individually for each model """
    R_ss = np.zeros((y.shape[0], y.shape[0], K), dtype=np.float64)
    A, B, C = ([], [], [])
    for m in range(K):
        m_results = ssm_array[m].m_estimate(y=y, x_t_n=x_t_n_all[m], P_t_n=P_t_n_all[m], P_t_tmin1_n=P_t_tmin1_n_all[m],
                                            h_t=h_t_m[m, :], priors=priors_all[m],
                                            keep_param=keep_param, return_dict=True)
        # Store sum variables for joint estimations
        R_ss[:, :, m] = m_results['R_ss']
        A.append(m_results['A'])
        B.append(m_results['B'])
        C.append(m_results['C'])

    """ Joint estimation of parameters """
    # Update observation noise R shared across models
    if shared_R and 'R' not in keep_param:
        R_ss = R_ss.sum(axis=2)
        # noinspection PyProtectedMember
        _ = [x._m_update_r(R_ss=R_ss, T=T, priors=priors_all[0][0]) for x in ssm_array]

    # Update F and Q matrices shared across models
    if shared_comp and not ('F' in keep_param and 'Q' in keep_param):
        shared_comp = (shared_comp,) if not isinstance(shared_comp, tuple) else shared_comp
        assert all([k == K for k in [len(x) for x in shared_comp]]), \
            'Index of component shared is not specified for all models. Use index=None if a model does not share.'

        for j in range(len(shared_comp)):  # iterate through each independent component that is shared
            # Obtain the component indices for a shared component in each model
            shared_comp_indices: list = shared_comp[j]  # type: ignore

            # Identify which models are joining
            joint_models = [x is not None for x in shared_comp_indices]

            # Total T is multiplied by the number of models sharing the components
            shared_T = T * sum(joint_models)

            # Get the state dimension of the current component
            first_model_index = np.argmax(joint_models)
            nstate = ssm_array[first_model_index].comp_nstates[shared_comp_indices[first_model_index]]

            # Create temporary trellis variables
            update_index_store = [(None, None)] * K
            A_tmp = np.zeros((nstate, nstate, K), dtype=np.float64)
            B_tmp = np.zeros((nstate, nstate, K), dtype=np.float64)
            C_tmp = np.zeros((nstate, nstate, K), dtype=np.float64)

            # Accumulate the sums of squares for joint estimation
            for m in range(K):
                comp_index = shared_comp_indices[m]
                if comp_index is not None:
                    # Identify the state indices for the component
                    start_idx = sum(ssm_array[m].comp_nstates[:comp_index])
                    end_idx = sum(ssm_array[m].comp_nstates[:comp_index + 1])
                    update_index_store[m] = (start_idx, end_idx)

                    # Accumulate sums of squares for the shared component
                    A_tmp[:, :, m] = A[m][start_idx:end_idx, start_idx:end_idx]
                    B_tmp[:, :, m] = B[m][start_idx:end_idx, start_idx:end_idx]
                    C_tmp[:, :, m] = C[m][start_idx:end_idx, start_idx:end_idx]

            A_tmp = A_tmp.sum(axis=2)
            B_tmp = B_tmp.sum(axis=2)
            C_tmp = C_tmp.sum(axis=2)

            # Assume all components being shared have the same priors specified
            comp_priors = priors_all[first_model_index][shared_comp_indices[first_model_index]]

            # Call component specific _m_update_<param> methods
            comp = ssm_array[first_model_index].components[shared_comp_indices[first_model_index]]
            if 'F' not in keep_param:
                # noinspection PyProtectedMember
                F = comp._m_update_f(A=A_tmp, B=B_tmp, priors=comp_priors)
            else:
                # assume that all models already have the same F entries for the shared component
                start_idx, end_idx = update_index_store[first_model_index]
                F = ssm_array[first_model_index].F[start_idx:end_idx, start_idx:end_idx]

            if 'Q' not in keep_param:
                # noinspection PyProtectedMember
                Q = comp._m_update_q(A=A_tmp, B=B_tmp, C=C_tmp, T=shared_T, F=F, priors=comp_priors)

            # Distribute to all models sharing the component
            for m in range(K):
                comp_index = shared_comp_indices[m]
                if comp_index is not None:
                    start_idx, end_idx = update_index_store[m]
                    if 'F' not in keep_param:
                        ssm_array[m].F[start_idx:end_idx, start_idx:end_idx] = F
                    if 'Q' not in keep_param:
                        # noinspection PyUnboundLocalVariable
                        ssm_array[m].Q[start_idx:end_idx, start_idx:end_idx] = Q

        for m in range(K):  # make sure instance attributes reflect the updated F and Q
            ssm_array[m].update_comp_param()


def _m2(h_t_m, h_t_tmin1_m):
    """ ML estimation of HMM parameters for the M.2 step """
    # M step - MLE of parameters {A, p1}
    p1 = h_t_m[:, 0]  # updated p1
    edges = h_t_tmin1_m.sum(axis=2)  # sum over time points
    A = edges / edges.sum(axis=0)  # updated A
    return A, p1


def _plot_e(h_t_m, q_t_m, E_iter):
    """ Plot the h_t_m and q_t_m for two models in E step """
    import matplotlib.pyplot as plt

    plt.subplots_adjust(hspace=0.5)
    fig = plt.figure()
    ax1 = plt.subplot(2, 1, 1)
    ax1.clear()
    ax1.plot(h_t_m[0, :])
    ax1.plot(h_t_m[1, :])
    ax1.set_title('h_t_m: iter ' + str(E_iter))
    ax1.set_ylim([-0.1, 1.1])

    ax2 = plt.subplot(2, 1, 2)
    ax2.clear()
    ax2.plot(q_t_m[0, :])
    ax2.plot(q_t_m[1, :])
    ax2.set_title('q_t_m: iter ' + str(E_iter))

    plt.pause(0.05)
    return fig



"""
Author: Mingjian He <mh1@stanford.edu>

Testing functions for basic state-space models in somata/basic_models
"""

from codetiming import Timer
from test_load_data import _load_data  # type: ignore

import os
os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"  # suppress a duplicate OpenMP library error with pytorch on Mac osx-64


def test_ssm():
    from somata.basic_models import StateSpaceModel as Ssm

    input_data = _load_data("kalman_inputs.mat")
    input_data.pop('R_weights')
    s1 = Ssm(components='Osc', **input_data)
    print(s1)

    with Timer():
        _ = s1.kalman_filt_smooth()
        _ = s1.dejong_filt_smooth()
        _ = Ssm.par_kalman(s1, method='kalman')
        _ = Ssm.par_kalman(s1, method='dejong')
        s1.m_estimate(**s1.dejong_filt_smooth(EM=True))

    _ = Ssm()


def test_gen():
    from somata.basic_models import GeneralSSModel as Gen

    input_data = _load_data("kalman_inputs.mat")
    input_data.pop('R_weights')
    g1 = Gen(**input_data)
    print(g1)

    with Timer():
        _ = g1.kalman_filt_smooth()
        _ = g1.dejong_filt_smooth()
        _ = Gen.par_kalman(g1, method='kalman')
        _ = Gen.par_kalman(g1, method='dejong')
        g1.m_estimate(**g1.dejong_filt_smooth(EM=True))

    _ = Gen()


def test_osc():
    from somata.basic_models import OscillatorModel as Osc

    input_data = _load_data("kalman_inputs.mat")
    input_data.pop('R_weights')
    o1 = Osc(**input_data, Fs=100)
    print(o1)

    with Timer():
        _ = o1.kalman_filt_smooth()
        _ = o1.dejong_filt_smooth()
        _ = Osc.par_kalman(o1, method='kalman')
        _ = Osc.par_kalman(o1, method='dejong')
        o1.m_estimate(**o1.dejong_filt_smooth(EM=True))

    _ = Osc()
    o2 = Osc(freq=[1, 10, 15], Fs=100, add_dc=True)
    o1.append(o2)
    _ = Osc(freq=5, Fs=80)
    _ = Osc(w=0.1)
    _ = Osc(add_dc=True)


def test_arn():
    from somata.basic_models import AutoRegModel as Arn

    input_data = _load_data("kalman_inputs.mat")
    input_data.pop('R_weights')
    a1 = Arn(coeff=0.95, R=input_data['R'], y=input_data['y'])
    print(a1)

    with Timer():
        _ = a1.kalman_filt_smooth()
        _ = a1.dejong_filt_smooth()
        _ = Arn.par_kalman(a1, method='kalman')
        _ = Arn.par_kalman(a1, method='dejong')
        a1.m_estimate(**a1.dejong_filt_smooth(EM=True))

    _ = Arn()
    a2 = Arn(coeff=[0.99, 0.5, 0.1])
    a1.append(a2)
    _ = Arn(F=a1.F)


if __name__ == "__main__":
    test_ssm()
    test_gen()
    test_osc()
    test_arn()
    print('SOMATA class tests finished without exception.')



"""
Author: Mingjian He <mh1@stanford.edu>

traditional module contains traditional methods for switching inference
"""

from somata.basic_models import StateSpaceModel as Ssm
from somata.exact_inference import forward_backward, logdet, inverse
import numpy as np


def switching(ssm_array, y=None, method='static', dwell_prob=0.99, A=None,
              future_steps=0, fix_prior=False, mimic1991=False, HMM_smooth=None):
    """
    Traditional switching inference functions.
    This is a collection of multiple traditional switching state-space
    model inference algorithms. There are a few different approaches,
    each associated with different input parameter options. All the
    methods output posterior model probabilities in the end. The
    references associated with each method will be provided above the
    implementation.

    Inputs:
    :param ssm_array: an array of Ssm class objects or a single object with stacked models
    :param y: observed data
    :param method: which traditional switching inference method to use.
                   Options are:
                    - 'static' - static multiple model using Bayes rule
                    - 'gpb1' - first-order generalized pseudo-Bayesian estimator
                    - 'gpb2' - second-order generalized pseudo-Bayesian estimator
                    - 'imm' - interacting multiple model algorithm
                    - '1991' - Shumway & Stoffer 1991 modified Kalman filtering method
                    - 'a parp' - Parallel method: predicted density for SSM with HMM filtering
                    - 'a pari' - Parallel method: interpolated density for SSM with HMM filtering
                    - 'ab parp' - Parallel method: predicted density for SSM with HMM smoothing
                    - 'ab pari' - Parallel method: interpolated density for SSM with HMM smoothing
                    - 'a pars' and 'ab pars': testing methods, do not use
    :param dwell_prob: dwell probability for HMM transitions to initialize A
    :param A: HMM transition matrix, if provided will override dwell_prob
    :param future_steps: A heuristic pseudo smoothing technique provided for the 1991 method
    :param fix_prior: parameter specific to parallel methods, whether to fix prior in HMM
                      filtering step, if not False will remove temporal continuity and MC model
    :param mimic1991: parameter specific to parallel methods, whether to use fy_t as an
                      approximation to fy_j(t|t-1) in the 1991 method and follow with the
                      same form of model probability filtering
    :param HMM_smooth: an option for all traditional methods (except parallel methods) to
                       apply the forward-backward algorithm to fy_t to obtain filtered
                       and smoothed model probabilities. None -> no additional processing,
                       'a' -> filtering, 'ab' -> smoothing. Parallel methods will set the
                       option automatically depending on what version is specified.

    Outputs:
    :returns:
        - Mprob: posterior model probabilities
        - fy_t: conditional density of y given t=1...t-1
    """
    ssm_array, K, T = Ssm.setup_array(ssm_array, y=y)

    # Initial guess of transition probability matrix A for HMM
    A = abs(np.eye(K, dtype=np.float64) -
            np.ones((K, K), dtype=np.float64) * (1-dwell_prob) / (K-1)) if A is None else A

    """
    Choose among the various switching state-space methods:
        - 'static' is similar to 'a parp'
        - 'gpb1'   is similar to '1991'
        - 'gpb2'   is similar to 'imm'
    """
    if method == 'static':
        Mprob, fy_t = _static(ssm_array, K, T)
    elif method == 'gpb1':
        Mprob, fy_t = _gpb1(ssm_array, K, T, A)
    elif method == 'gpb2':
        Mprob, fy_t = _gpb2(ssm_array, K, T, A)
    elif method == 'imm':
        Mprob, fy_t = _imm(ssm_array, K, T, A)
    elif method == '1991':
        Mprob, fy_t = _1991(ssm_array, K, T, A, future_steps=future_steps)
    elif 'par' in method:
        Mprob, fy_t, HMM_smooth = _parallel(ssm_array, K, T, A, method, fix_prior=fix_prior, mimic1991=mimic1991)
    else:
        raise Exception('Unrecognized switching inference method.')

    # Truncate the first time point to output vectors t=1 -> t=T
    Mprob = Mprob[:, 1:] if 'par' not in method else Mprob

    # Apply forward-backward algorithm to fy_t if specified
    if HMM_smooth is not None:
        if HMM_smooth == 'a':  # alpha filtering
            Mprob, *_ = forward_backward(A, fy_t)
        elif HMM_smooth == 'ab':  # alpha-beta smoothing
            _, Mprob, *_ = forward_backward(A, fy_t)

    return Mprob, fy_t


def _static(ssm_array, K, T):
    """
    This is the naive static multiple model derived using
    a direct application of Bayes rule. It essentially
    uses predicted conditional density and runs an update
    on the prior for each time step. This method is
    obsolete now given the later development of dynamic
    methods, however with an ad-hoc modification of lower
    bounding model probability, it is still used in the
    literature.

    Reference:
        Fabri, S., & Kadirkamanathan, V. (2001). Functional
        adaptive control: an intelligent systems approach.
        Springer Science & Business Media.
    """
    # Parallel run Kalman filtering on the static models
    _, _, _, _, _, _, _, x_t_tmin1_all, P_t_tmin1_all, _ = Ssm.par_kalman(ssm_array, method='kalman')

    # Compute the conditional density (aka. predicted likelihood function of y)
    fy_t = _compute_fy_t(ssm_array, x_t_tmin1_all, P_t_tmin1_all)

    # Initialize posterior model probability
    Mprob = np.zeros((K, T+1), dtype=np.float64)  # (index 1 corresponds to t=0)
    Mprob[:, 0] = np.ones(K, dtype=np.float64) / K

    # Iterate through the time steps
    for ii in range(1, T+1):  # t=1 -> t=T
        bayes_num = fy_t[:, ii-1] * Mprob[:, ii-1]
        Mprob[:, ii] = bayes_num / bayes_num.sum()  # direct application of Bayes rule

        # An ad-hoc solution of lower bounding model probability
        lower_bound = 1e-2
        if any(Mprob[:, ii] < lower_bound):
            Mprob[Mprob[:, ii] < lower_bound, ii] = lower_bound  # lower bound
            Mprob[:, ii] = Mprob[:, ii] / Mprob[:, ii].sum()  # re-normalize

    return Mprob, fy_t


def _gpb1(ssm_array, K, T, A, y=None):
    """
    This implements the generalized pseudo-Bayesian
    estimator of switching multiple models. The name
    comes from the fact that we combine at the end of
    each filtering + update step the K separate Gaussian
    distributions into a single approximating Gaussian.

    The merging of Gaussian is simply weighted averaging
    of the first two moments of the filtered state
    estimates using a pseudo-Bayesian update rule. It is
    not strictly Bayesian because the prior involves an
    averaging over transition matrix of discrete HMM.

    Note that this method is intimately related to the
    S&S 1991 approach. The only difference is that S&S
    1991 takes the weighted average of innovations while
    gpb1 takes the weighted average of filtered state
    estimates. The former is exact solution derived under
    the assumption that only the observation matrices
    switch, while the latter is approximating a mixture
    of K Gaussian by merging their first two moments.

    If K models have the same model parameters (1991),
    at every iteration, the two methods produce the same
    first moment for the merged filtered estimate, but
    gpb1 has an additional covariance matrix coming from
    the deviation of mean of each of K Gaussian from the
    mean of the merged Gaussian. Similarly, one can use
    the 1991 method to implement gpb1 by using augmented
    hidden states and making F and Q block diagonal.

    Reference:
        Gordon, K., & Smith, A. F. M. (1990). Modeling and
        monitoring biomedical time series. Journal of the
        American Statistical Association, 85(410), 328-337.
    """
    # gpb1 requires all the models to have the same dimensions
    # of the hidden states in order to merge Gaussian - this
    # is verified here
    first_ssm = ssm_array[0]
    for m in range(1, K):
        assert first_ssm.mu0.shape[0] == ssm_array[m].G.shape[1], 'Model hidden state dimensions are inconsistent.'

    # Array dimensions
    y = first_ssm.y if y is None else y
    p = first_ssm.nstate
    q = first_ssm.nchannel
    approach = 'svd' if q >= 5 else 'gaussian'
    qlog2pi = q * np.log(2 * np.pi)

    # Forward filtering trellis variables
    x_t_tmin1 = np.zeros((p, K, T+1), dtype=np.float64)  # (index 1 corresponds to t=0, etc.)
    P_t_tmin1 = np.zeros((p, p, K, T+1), dtype=np.float64)
    K_t = np.zeros((p, q, K, T+1), dtype=np.float64)
    x_t_t = np.zeros((p, T+1), dtype=np.float64)
    P_t_t = np.zeros((p, p, T+1), dtype=np.float64)
    Mprob = np.zeros((K, T+1), dtype=np.float64)
    fy_t = np.zeros((K, T), dtype=np.float64)  # conditional density of y given t=1...t-1 (index 1 corresponds to t=1)

    # Initial model probability at t=0
    Mprob[:, 0] = np.ones(K, dtype=np.float64) / K

    # Initialize x_0_0
    x_t_t[:, 0] = np.array([ssm_array[m].mu0 * Mprob[m, 0] for m in range(K)]).sum(axis=0)[:, 0]

    # Initialize P_0_0
    for m in range(K):
        P_t_t[:, :, 0] += (ssm_array[m].S0 +
                           (ssm_array[m].mu0 - x_t_t[:, 0][:, None]) @
                           (ssm_array[m].mu0 - x_t_t[:, 0][:, None]).T) * Mprob[m, 0]

    # Initialize log likelihood of filtering for each model
    logL = np.zeros(K, dtype=np.float64)

    # Forward iterations
    for ii in range(1, T+1):  # t=1 -> t=T
        # Store filtered estimates for each model to later combine
        # as one Gaussian at the end of the step
        x_t_t_all = np.zeros((p, K), dtype=np.float64)
        P_t_t_all = np.zeros((p, p, K), dtype=np.float64)

        # Separate filter for each model
        for m in range(K):
            # Current model parameters
            F = ssm_array[m].F
            Q = ssm_array[m].Q
            G = ssm_array[m].G
            R = ssm_array[m].R

            # One-step prediction
            x_t_tmin1[:, m, ii] = F @ x_t_t[:, ii-1]
            P_t_tmin1[:, :, m, ii] = F @ P_t_t[:, :, ii-1] @ F.T + Q
            Sigma = G @ P_t_tmin1[:, :, m, ii] @ G.T + R
            K_t[:, :, m, ii] = P_t_tmin1[:, :, m, ii] @ G.T @ inverse(Sigma, approach=approach)

            # Current time step Gaussian pdf of y_t
            e_t = y[:, ii-1] - G @ x_t_tmin1[:, m, ii]
            fy_t[m, ii-1] = np.exp(-(qlog2pi + logdet(Sigma) + e_t.T @ inverse(Sigma, approach=approach) @ e_t) / 2)

            # Innovation form of the log likelihood for the current model
            logL[m] += np.log(fy_t[m, ii-1])

            # Update equation to get filtered estimates for each model
            x_t_t_all[:, m] = x_t_tmin1[:, m, ii] + K_t[:, :, m, ii] @ e_t
            P_t_t_all[:, :, m] = P_t_tmin1[:, :, m, ii] - K_t[:, :, m, ii] @ G @ P_t_tmin1[:, :, m, ii]

        # Update model probability from last step
        Mprob_prior = A @ Mprob[:, ii-1]  # prior model probability
        bayes_num = fy_t[:, ii-1] * Mprob_prior
        Mprob[:, ii] = bayes_num / bayes_num.sum()  # filtered model probability

        # Merge the K filtered Gaussian into one Gaussian
        x_t_t[:, ii] = (x_t_t_all * Mprob[:, ii].T).sum(axis=1)
        for m in range(K):
            P_t_t[:, :, ii] += (P_t_t_all[:, :, m] +
                                (x_t_t_all[:, m] - x_t_t[:, ii])[:, None] @
                                (x_t_t_all[:, m] - x_t_t[:, ii])[:, None].T) * Mprob[m, ii]

    return Mprob, fy_t


def _gpb2(ssm_array, K, T, A, y=None):
    """
    This implements the second-order generalized pseudo
    Bayesian approach as described in Bar-Shalom & Li
    1993. The method is an extension of gpb1 but differs
    by maintaining K different hidden state estimate
    distributions similar to the IMM method.

    The computations are similar to IMM but have K^2
    number of filters. Merging of K Gaussian into single
    approximating Gaussian is done for each of the model
    at the end of each time step. This is different from
    IMM, which mixes distributions at the beginning of
    every step which requires K filters instead of K^2.

    It can be shown that IMM of second order is identical
    to gpb2, therefore this method also serves as a
    second order extension of the IMM method.

    Reference:
        Bar-Shalom, Y., Li, X. R., & Kirubarajan, T. (2004).
        Estimation with applications to tracking and
        navigation: theory algorithms and software. John
        Wiley & Sons.
    """
    # gpb2 requires all the models to have the same dimensions
    # of the hidden states in order to merge Gaussian - this
    # is verified here
    first_ssm = ssm_array[0]
    for m in range(1, K):
        assert first_ssm.mu0.shape[0] == ssm_array[m].G.shape[1], 'Model hidden state dimensions are inconsistent.'

    # Array dimensions
    y = first_ssm.y if y is None else y
    p = first_ssm.nstate
    q = first_ssm.nchannel
    approach = 'svd' if q >= 5 else 'gaussian'
    qlog2pi = q * np.log(2 * np.pi)

    # Forward filtering trellis variables
    x_t_tmin1 = np.zeros((p, K, K, T+1), dtype=np.float64)  # (index 1 corresponds to t=0, etc.)
    P_t_tmin1 = np.zeros((p, p, K, K, T+1), dtype=np.float64)
    K_t = np.zeros((p, q, K, K, T+1), dtype=np.float64)
    x_t_t = np.zeros((p, K, T+1), dtype=np.float64)
    P_t_t = np.zeros((p, p, K, T+1), dtype=np.float64)
    Mprob = np.zeros((K, T+1), dtype=np.float64)
    fy_t = np.zeros((K, T), dtype=np.float64)  # conditional density of y given t=1...t-1 (index 1 corresponds to t=1)
    fy_t_all = np.zeros((K, K, T), dtype=np.float64)  # used for model probability update equation

    # Initial model probability at t=0
    Mprob[:, 0] = np.ones(K, dtype=np.float64) / K

    # Initialize x_0_0 and P_0_0
    for m in range(K):
        x_t_t[:, m, 0] = ssm_array[m].mu0[:, 0]
        P_t_t[:, :, m, 0] = ssm_array[m].S0

    # Initialize log likelihood of filtering for each model
    logL = np.zeros(K, dtype=np.float64)

    # Forward iterations
    for ii in range(1, T+1):  # t=1 -> t=T
        # Separate filter for each model
        for m in range(K):
            # Store filtered estimates for each model to later combine
            # as one Gaussian at the end of the step
            x_t_t_all = np.zeros((p, K), dtype=np.float64)
            P_t_t_all = np.zeros((p, p, K), dtype=np.float64)

            # Current model parameters
            F = ssm_array[m].F  # in state m at current time step
            Q = ssm_array[m].Q
            G = ssm_array[m].G
            R = ssm_array[m].R

            # Evaluate each model separately
            for j in range(K):
                # One-step prediction
                x_t_tmin1[:, m, j, ii] = F @ x_t_t[:, j, ii-1]  # in state j at last time step
                P_t_tmin1[:, :, m, j, ii] = F @ P_t_t[:, :, j, ii-1] @ F.T + Q
                Sigma = G @ P_t_tmin1[:, :, m, j, ii] @ G.T + R
                K_t[:, :, m, j, ii] = P_t_tmin1[:, :, m, j, ii] @ G.T @ inverse(Sigma, approach=approach)

                # Current time step Gaussian pdf of y_t
                e_t = y[:, ii-1] - G @ x_t_tmin1[:, m, j, ii]
                fy_t_all[m, j, ii-1] = np.exp(-(qlog2pi + logdet(Sigma) + e_t.T @ inverse(Sigma,
                                                                                          approach=approach) @ e_t) / 2)

                # Update equation to get filtered estimates
                x_t_t_all[:, j] = x_t_tmin1[:, m, j, ii] + K_t[:, :, m, j, ii] @ e_t
                P_t_t_all[:, :, j] = P_t_tmin1[:, :, m, j, ii] - K_t[:, :, m, j, ii] @ G @ P_t_tmin1[:, :, m, j, ii]

            # Compute the merging probability
            bayes_num = fy_t_all[m, :, ii-1] * A[m, :] * Mprob[:, ii-1].T
            Mprob_merge = bayes_num / bayes_num.sum()

            # Merge the K Gaussian into one Gaussian for each model
            x_t_t[:, m, ii] = (x_t_t_all * Mprob_merge).sum(axis=1)
            for j in range(K):
                P_t_t[:, :, m, ii] += (P_t_t_all[:, :, j] +
                                       (x_t_t_all[:, j] - x_t_t[:, m, ii])[:, None] @
                                       (x_t_t_all[:, j] - x_t_t[:, m, ii])[:, None].T) * Mprob_merge[j]

            # Weighted averaging of conditional density to get fy_t
            fy_t[m, ii-1] = fy_t_all[m, :, ii-1] @ Mprob_merge

            # Innovation form of the log likelihood for the current model
            logL[m] += np.log(fy_t[m, ii-1])

        # Update model probability from last step
        bayes_mat = fy_t_all[:, :, ii-1] * A * Mprob[:, ii-1].T
        Mprob[:, ii] = bayes_mat.sum(axis=1) / bayes_mat.sum()

    return Mprob, fy_t


def _imm(ssm_array, K, T, A, y=None):
    """
    This implements the interacting multiple model (IMM)
    approach as sketched out in Bar-Shalom & Li 1993. The
    method is distinguished by the key feature of mixing
    Gaussian at beginning of a step in Kalman filtering
    for each of the K models. Therefore, this method
    maintains K different hidden state estimate
    distributions, which interact at the beginning of
    every step using mixing probabilities that are
    model probabilities at time step t-1 conditioned on
    being at a specific state at time step t. Note that
    this is different from the posterior model
    probabilities used for mixing filtered estimates
    (gpb1) or innovations (1991), which occur at the end
    of a time step in Kalman filtering.

    Reference:
        Bar-Shalom, Y., Li, X. R., & Kirubarajan, T. (2004).
        Estimation with applications to tracking and
        navigation: theory algorithms and software. John
        Wiley & Sons.
    """
    # imm requires all the models to have the same dimensions
    # of the hidden states in order to merge Gaussian - this
    # is verified here
    first_ssm = ssm_array[0]
    for m in range(1, K):
        assert first_ssm.mu0.shape[0] == ssm_array[m].G.shape[1], 'Model hidden state dimensions are inconsistent.'

    # Array dimensions
    y = first_ssm.y if y is None else y
    p = first_ssm.nstate
    q = first_ssm.nchannel
    approach = 'svd' if q >= 5 else 'gaussian'
    qlog2pi = q * np.log(2 * np.pi)

    # Forward filtering trellis variables
    x_t_tmin1 = np.zeros((p, K, T+1), dtype=np.float64)  # (index 1 corresponds to t=0, etc.)
    P_t_tmin1 = np.zeros((p, p, K, T+1), dtype=np.float64)
    K_t = np.zeros((p, q, K, T+1), dtype=np.float64)
    x_t_t = np.zeros((p, K, T+1), dtype=np.float64)
    P_t_t = np.zeros((p, p, K, T+1), dtype=np.float64)
    Mprob = np.zeros((K, T+1), dtype=np.float64)
    fy_t = np.zeros((K, T), dtype=np.float64)  # conditional density of y given t=1...t-1 (index 1 corresponds to t=1)

    # Initial model probability at t=0
    Mprob[:, 0] = np.ones(K, dtype=np.float64) / K

    # Initialize x_0_0 and P_0_0
    for m in range(K):
        x_t_t[:, m, 0] = ssm_array[m].mu0[:, 0]
        P_t_t[:, :, m, 0] = ssm_array[m].S0

    # Initialize log likelihood of filtering for each model
    logL = np.zeros(K, dtype=np.float64)

    # Forward iterations
    for ii in range(1, T+1):  # t=1 -> t=T
        # Compute the mixing probability
        bayes_num = A * Mprob[:, ii-1].T
        # each row is the mixing probability of other models for model m
        Mprob_mix = bayes_num / bayes_num.sum(axis=1)[:, None]

        # Separate filter for each model
        for m in range(K):
            # Current model parameters
            F = ssm_array[m].F
            Q = ssm_array[m].Q
            G = ssm_array[m].G
            R = ssm_array[m].R

            # Mix the last step hidden states from K models
            x0_t_t = (x_t_t[:, :, ii-1] * Mprob_mix[m, :]).sum(axis=1)
            P0_t_t = np.zeros((p, p), dtype=np.float64)
            for j in range(K):
                P0_t_t += (P_t_t[:, :, j, ii-1] +
                           (x_t_t[:, j, ii-1] - x0_t_t)[:, None] @
                           (x_t_t[:, j, ii-1] - x0_t_t)[:, None].T) * Mprob_mix[m, j]

            # One-step prediction using the merged Gaussian
            x_t_tmin1[:, m, ii] = F @ x0_t_t
            P_t_tmin1[:, :, m, ii] = F @ P0_t_t @ F.T + Q
            Sigma = G @ P_t_tmin1[:, :, m, ii] @ G.T + R
            K_t[:, :, m, ii] = P_t_tmin1[:, :, m, ii] @ G.T @ inverse(Sigma, approach=approach)

            # Current time step Gaussian pdf of y_t
            e_t = y[:, ii-1] - G @ x_t_tmin1[:, m, ii]
            fy_t[m, ii-1] = np.exp(-(qlog2pi + logdet(Sigma) + e_t.T @ inverse(Sigma, approach=approach) @ e_t) / 2)

            # Innovation form of the log likelihood for the current model
            logL[m] += np.log(fy_t[m, ii-1])

            # Update equation to get filtered estimates for each model
            x_t_t[:, m, ii] = x_t_tmin1[:, m, ii] + K_t[:, :, m, ii] @ e_t
            P_t_t[:, :, m, ii] = P_t_tmin1[:, :, m, ii] - K_t[:, :, m, ii] @ G @ P_t_tmin1[:, :, m, ii]

        # Update model probability from last step
        Mprob_prior = (A * Mprob[:, ii-1].T).sum(axis=1)
        bayes_num = fy_t[:, ii-1] * Mprob_prior
        Mprob[:, ii] = bayes_num / bayes_num.sum()  # filtered model probability

    return Mprob, fy_t


def _1991(ssm_array, K, T, A, y=None, future_steps=0):
    """
    This implements the original Shumway & Stoffer 1991
    modified Kalman filtering approach for switching
    state space models. A distinguishing feature of this
    method from the gpb1 method is that only the
    observation matrices switch in S&S 1991, therefore
    the Kalman filtering equations are exact by weighting
    the innovations instead of merging Gaussian as
    approximation. Thus, unless using pseudo-EM to learn
    the SSM parameters (in S&S 1991 Appendix), this
    method does not involve approximating Gaussian like
    all the other traditional switching inference methods.

    Reference:
        Shumway, R. H., & Stoffer, D. S. (1991). Dynamic
        linear models with switching. Journal of the American
        Statistical Association, 86(415), 763-769.

    There is an optional pseudo-smoothing available by
    specifying <future_steps> to be greater than 0. It
    is a heuristic modification of the 1991 method by
    replacing the conditional density of y with the
    summation of the conditional density of y under each
    alternative model after filtering into the future for
    <future_steps> steps. This is not strictly derivable
    under Bayes rule and therefore only serves as a
    heuristic solution.
    """
    # 1991 requires all the models to have the same properties
    # except the observation matrices - this is verified here
    first_ssm = ssm_array[0]
    for m in range(1, K):
        assert (first_ssm.F == ssm_array[m].F).all(), 'Models have different F matrices.'
        assert (first_ssm.Q == ssm_array[m].Q).all(), 'Models have different Q matrices.'
        assert (first_ssm.mu0 == ssm_array[m].mu0).all(), 'Models have different mu0 vectors.'
        assert (first_ssm.S0 == ssm_array[m].S0).all(), 'Models have different S0 matrices.'
        assert (first_ssm.R == ssm_array[m].R).all(), 'Models have different R matrices.'
        assert first_ssm.mu0.shape[0] == ssm_array[m].G.shape[1], 'Model hidden state dimensions are inconsistent.'

    # Array dimensions
    y = first_ssm.y if y is None else y
    p = first_ssm.nstate
    q = first_ssm.nchannel
    approach = 'svd' if q >= 5 else 'gaussian'
    qlog2pi = q * np.log(2 * np.pi)

    # Forward filtering trellis variables
    x_t_tmin1 = np.zeros((p, T+1), dtype=np.float64)  # (index 1 corresponds to t=0, etc.)
    P_t_tmin1 = np.zeros((p, p, T+1), dtype=np.float64)
    K_t = np.zeros((p, q, K, T+1), dtype=np.float64)
    x_t_t = np.zeros((p, T+1), dtype=np.float64)
    P_t_t = np.zeros((p, p, T+1), dtype=np.float64)
    Mprob = np.zeros((K, T+1), dtype=np.float64)
    fy_t = np.zeros((K, T), dtype=np.float64)  # conditional density of y given t=1...t-1 (index 1 corresponds to t=1)

    # Initial model probability at t=0
    Mprob[:, 0] = np.ones(K, dtype=np.float64) / K  # pi_0_0

    # Initialize hidden states at t=0
    x_t_t[:, 0] = first_ssm.mu0[:, 0]  # x_0_0
    P_t_t[:, :, 0] = first_ssm.S0  # P_0_0

    # Initialize log likelihood of filtering for each model
    logL = np.zeros(K, dtype=np.float64)

    # Set up model parameters
    F = first_ssm.F
    Q = first_ssm.Q
    R = first_ssm.R

    # Forward iterations
    for ii in range(1, T+1):  # t=1 -> t=T
        # One-step prediction
        x_t_tmin1[:, ii] = F @ x_t_t[:, ii-1]
        P_t_tmin1[:, :, ii] = F @ P_t_t[:, :, ii-1] @ F.T + Q

        # Compute innovations from K models
        fy_t_tmin1 = np.zeros((K, min(T+1-ii, future_steps+1)), dtype=np.float64)
        for m in range(K):
            G = ssm_array[m].G  # current model observation matrix
            Sigma = G @ P_t_tmin1[:, :, ii] @ G.T + R
            K_t[:, :, m, ii] = P_t_tmin1[:, :, ii] @ G.T @ inverse(Sigma, approach=approach)

            # Current time step Gaussian pdf of y_t
            e_t = y[:, ii-1] - G @ x_t_tmin1[:, ii]
            fy_t_tmin1[m, 0] = np.exp(-(qlog2pi + logdet(Sigma) + e_t.T @ inverse(Sigma, approach=approach) @ e_t) / 2)

            # Innovation form of the log likelihood for the current model
            logL[m] += np.log(fy_t_tmin1[m, 0])

            # Filter into the future if future_steps > 0
            temp_x_t_tmin1 = np.zeros((p, future_steps+1), dtype=np.float64)
            temp_x_t_tmin1[:, 0] = x_t_tmin1[:, ii]
            temp_P_t_tmin1 = np.zeros((p, p, future_steps+1), dtype=np.float64)
            temp_P_t_tmin1[:, :, 0] = P_t_tmin1[:, :, ii]
            temp_K_t = np.zeros((p, q, future_steps+1), dtype=np.float64)
            temp_K_t[:, :, 0] = K_t[:, :, m, ii]
            temp_x_t_t = np.zeros((p, future_steps+1), dtype=np.float64)
            temp_P_t_t = np.zeros((p, p, future_steps+1), dtype=np.float64)

            # Filter forward in the current model for <future_steps> steps
            for k in range(1, min(T+1-ii, future_steps+1)):
                # Finish off the filtering for the last time step
                temp_x_t_t[:, k-1] = temp_x_t_tmin1[:, k-1] + temp_K_t[:, :, k-1] @ (y[:, ii-1+k-1] -
                                                                                     G @ temp_x_t_tmin1[:, k-1])
                temp_P_t_t[:, :, k-1] = temp_P_t_tmin1[:, :, k-1] - temp_K_t[:, :, k-1] @ G @ temp_P_t_tmin1[:, :, k-1]

                # Compute the prediction for the current step
                temp_x_t_tmin1[:, k] = F @ temp_x_t_t[:, k-1]
                temp_P_t_tmin1[:, :, k] = F @ temp_P_t_t[:, :, k-1] @ F.T + Q
                temp_Sigma = G @ temp_P_t_tmin1[:, :, k] @ G.T + R
                temp_K_t[:, :, k] = temp_P_t_tmin1[:, :, k] @ G.T @ inverse(temp_Sigma, approach=approach)

                # Calculate the conditional density of f_y for the current step
                temp_e_t = y[:, ii-1+k] - G @ temp_x_t_tmin1[:, k]
                fy_t_tmin1[m, k] = np.exp(-(qlog2pi + logdet(temp_Sigma) +
                                            temp_e_t.T @ inverse(temp_Sigma, approach=approach) @ temp_e_t) / 2)

        # Calculate filtered model probabilities
        fy_t[:, ii-1] = fy_t_tmin1.sum(axis=1)  # sum conditional density after filtering <future_steps> with all models
        Mprob_prior = A @ Mprob[:, ii-1]  # prior model probability
        bayes_num = fy_t[:, ii-1] * Mprob_prior
        Mprob[:, ii] = bayes_num / bayes_num.sum()  # filtered model probability

        # Single joint filter by merging innovations from K models using filtered model probability
        x_update = np.zeros(p, dtype=np.float64)
        P_update = np.zeros((p, p), dtype=np.float64)
        for m in range(K):
            G = ssm_array[m].G  # current model observation matrix
            x_update += Mprob[m, ii] * K_t[:, :, m, ii] @ (y[:, ii-1] - G @ x_t_tmin1[:, ii])
            P_update += Mprob[m, ii] * (P_t_tmin1[:, :, ii] - K_t[:, :, m, ii] @ G @ P_t_tmin1[:, :, ii])
        x_t_t[:, ii] = x_t_tmin1[:, ii] + x_update
        P_t_t[:, :, ii] = P_update

    return Mprob, fy_t


def _parallel(ssm_array, K, T, A, method, fix_prior=False, mimic1991=False):
    """
    "Parallel model" approach treats the Gaussian process
    hidden states as known from the Kalman filtering and
    smoothing. The rest is then just the classical
    alpha-beta forward backward algorithm on a discrete
    state Markov Chain (hidden Markov model) that
    captures the switching state variable.

    Reference:
        He, M., Das, P., Hotan, G., & Purdon, P. L. (2023). Switching
        state-space modeling of neural signal dynamics. PLoS
        Computational Biology, 19(8), e1011395.

    The key assumption here is that after running
    Kalman filtering and smoothing of the parallel
    models, we use the predicted or interpolated density
    as approximations to the observation probability in
    the hidden Markov model and use forward backward
    algorithm to complete *approximate* inference.

    Two approximations are at play:
        1) Approximate conditional independence across y
        2) Approximate density with static history models

    Abbreviations:
        - 'a parp' = filtered MC with predicted Gaussian states (conditional independence is not necessary)
        - 'a pari' = filtered MC with interpolated Gaussian states (still non-causal)
        - 'ab parp' = smoothed MC with predicted Gaussian states
        - 'ab pari' = smoothed MC with interpolated Gaussian states (best option)
        - 'a pars' and 'ab pars': testing methods, do not use
    """
    """ SSM """
    # Run parallel kalman filtering and smoothing on an array of ssm objects,
    # then compute the conditional density of y at each time step
    #
    # Note that all the choices of hidden state estimates can be viewed as
    # approximations of fy_j(t|t-1) in the S&S 1991 approach if alpha-filtering
    # is used for HMM

    if 'parp' in method:  # predicted conditional density
        _, _, _, _, _, _, _, x_t_tmin1_all, P_t_tmin1_all, _ = Ssm.par_kalman(ssm_array, method='kalman')
        fy_t = _compute_fy_t(ssm_array, x_t_tmin1_all, P_t_tmin1_all)
    elif 'pars' in method:  # smoothed conditional density (do not use)
        x_t_n_all, P_t_n_all, *_ = Ssm.par_kalman(ssm_array, method='kalman')
        fy_t = _compute_fy_t(ssm_array, x_t_n_all, P_t_n_all)
    elif 'pari' in method:  # interpolated density using De Jong filtering and smoothing
        _, _, _, _, _, _, _, _, _, fy_t = Ssm.par_kalman(ssm_array, method='dejong', skip_interp=False)
        fy_t = np.vstack(fy_t)
    else:
        raise Exception('Unrecognized switching inference method.')

    """ HMM """
    # Force mimic1991 to be False if doing smoothing on HMM
    if 'ab' in method and mimic1991:
        raise Exception('Smoothing HMM cannot mimic the 1991 method.')

    if mimic1991:
        # If we try to use fy_t as an approximation of fy_j(t|t-1)
        # in the 1991 method, we need a slightly different
        # initialization of the alpha state at t=1, therefore we
        # cannot use te dedicated implementation of forward-backward
        # algorithm as the other traditional methods. We implement
        # the associated alpha filtering below.
        HMM_smooth = None

        if fix_prior:
            assert len(fix_prior) == K, 'Fix prior needs to provide the prior for each model.'  # type: ignore
            # A fixed predictor / prior lacks temporal continuity, not MC anymore
            predictor = fix_prior  # this is provided as an option to mimic the 1991 method

        norm_a = np.zeros((K, T), dtype=np.float64)  # (index 1 corresponds to t=0)
        norm_a[:, 0] = np.ones(K, dtype=np.float64) / K  # equivalent to pi_0_0 in the 1991 method
        for ii in range(1, T):  # t=1 -> t=T
            if not fix_prior:
                predictor = A @ norm_a[:, ii-1]  # equivalent to Mprob_prior in the 1991 method
            # noinspection PyUnboundLocalVariable
            a = fy_t[:, ii-1] * predictor
            norm_a[:, ii] = a / a.sum()

        # This is equivalent to the 1991 approach using custom hidden state
        # estimates to approximate fy_j(t|t-1) and doing the same filtering
        # estimates for discrete switching states
        Mprob = norm_a[:, 1:]  # Truncate the first time point to output vectors t=1 -> t=T
    else:
        # Invoke dedicated forward-backward algorithm with more conventional initialization
        Mprob = None
        HMM_smooth = method.split()[0]

    return Mprob, fy_t, HMM_smooth


def _compute_fy_t(ssm_array, x_t_all, P_t_all, y=None):
    """ Compute the conditional Gaussian density of P(y_t|x_t) """
    # Array dimensions
    y = ssm_array[0].y if y is None else y
    K = len(ssm_array)  # number of models
    q, T = y.shape  # number of channels and number of time points
    approach = 'svd' if q >= 5 else 'gaussian'
    qlog2pi = y.shape[0] * np.log(2 * np.pi)

    fy_t = np.zeros((K, T), dtype=np.float64)  # (index 1 corresponds to t=1)

    for m in range(K):
        x_t = x_t_all[m][:, 1:]  # (index 1 corresponds to be t=0, etc.)
        P_t = P_t_all[m][:, :, 1:]
        for ii in range(T):  # t=1 -> t=T
            Sigma = ssm_array[m].G @ P_t[:, :, ii] @ ssm_array[m].G.T + ssm_array[m].R
            e_t = y[:, ii] - ssm_array[m].G @ x_t[:, ii]
            fy_t[m, ii] = np.exp(-(qlog2pi + logdet(Sigma) + e_t.T @ inverse(Sigma, approach=approach) @ e_t) / 2)
    return fy_t



"""
Author: Mingjian He <mh1@stanford.edu>

exact_inference module contains basic dynamic programming algorithms used in SOMATA
"""

from .dp_func import forward_backward, baum_welch, viterbi, logdet, logdet_torch, \
    kalman, djkalman, djkalman_conv_torch, inverse, inverse_torch  # noqa: F401
from .em import run_em  # noqa: F401



"""
Author: Mingjian He <mh1@stanford.edu>

general helper functions used throughout SOMATA
"""


import numpy as np
from itertools import groupby
import scipy.signal as signal


def consecutive(data):
    """
    This function takes a list of values and returns a list of tuples.
    Each tuple contains the value, the start index, and the end index
    of a consecutive sequence of that value.
    Parameters
    ----------
    data : list or ndarray
        A list of values.
    Returns
    -------
    list
        A list of tuples. Each tuple contains the value, the start index,
        and the end index of a consecutive sequence of that value.
    Examples
    --------
    >> consecutive([1, 1, 1, 2, 2, 3, 3, 3, 3])
    [(1, 0, 2), (2, 3, 4), (3, 5, 8)]
    """
    data_group = [(v, sum(1 for _ in g)) for v, g in groupby(data)]
    vals, cons = zip(*data_group)

    start_inds = np.cumsum((0,) + cons[:-1])
    end_inds = start_inds + cons - 1

    return list(zip(vals, start_inds, end_inds))


def estimate_r(y, Fs, freq_cutoff, plot_freqz=False):
    """ Estimate the observation noise covariance R in a given time series """
    Fs_nq = Fs / 2  # Nyquist frequency
    wp = np.min([Fs_nq - 10, freq_cutoff])  # passband edge frequency in Hz
    ws = wp - 2  # stopband edge frequency in Hz
    N, Wn = signal.buttord(wp=wp, ws=ws, gpass=1, gstop=50, fs=Fs)  # select filter order
    sos = signal.butter(N, Wn, btype="high", output='sos', fs=Fs)  # highpass filter

    # Visualize the filter frequency response
    if plot_freqz:
        import matplotlib.pyplot as plt
        w, h = signal.sosfreqz(sos, fs=Fs)
        fig, ax = plt.subplots()
        ax.plot(w, 20 * np.log10(np.maximum(np.abs(h), 1e-3)))
        ax.set_title('Butterworth filter frequency response')
        ax.set_xlabel('Frequency [Hz]')
        ax.set_ylabel('Amplitude [dB]')
        ax.margins(0, 0.1)
        ax.grid(which='both', axis='both')
        ax.set_xlim([0, Fs_nq])

    # Apply the high pass filter and estimate observation noise covariance
    y_filt = signal.sosfiltfilt(sos, np.squeeze(y))
    R = np.cov(y_filt) * Fs_nq / (Fs_nq - wp)  # scale up based on wp

    return R


def wrapToPi(x):
    """ Wrap phase value in radian to -pi~pi """
    x_wrap = np.remainder(x, 2 * np.pi)
    mask = np.abs(x_wrap) > np.pi
    x_wrap[mask] -= 2 * np.pi * np.sign(x_wrap[mask])
    mask1 = x < 0
    mask2 = np.remainder(x, np.pi) == 0
    mask3 = np.remainder(x, 2 * np.pi) != 0
    x_wrap[mask1 & mask2 & mask3] -= 2 * np.pi
    return x_wrap



from .iter_osc import IterativeOscillatorModel  # class holding OscillatorModel class objects during iterative search  # noqa: F401, E501
from .decomp_osc import DecomposedOscillatorModel  # class holding OscillatorModel class objects during decomposition  # noqa: F401, E501



"""
Author: Mingjian He <mh1@stanford.edu>

Testing function for loading data in tests/test_*.py
"""

import pathlib
from importlib.util import find_spec
from scipy.io.matlab import loadmat


def test_anchor_func():
    """ Anchor function for locating this .py file """
    return


def get_test_path():
    """ Find the path to the directory containing the test_load_data.py file """
    f_path = pathlib.Path(test_anchor_func.__code__.co_filename)
    if f_path.is_file():
        return f_path.parents[0].resolve()
    else:
        somata_path = pathlib.Path(find_spec('somata').origin)  # find the package root __init__.py filepath
        return (somata_path.parents[1] / 'tests').resolve()  # 2-levels up


def _load_data(filename, return_path=False):
    """ Load data files for test functions """
    full_fn = get_test_path() / '_test_data' / filename
    if return_path:
        return full_fn
    else:
        data_dict = loadmat(full_fn)
        for elem in ('__header__', '__version__', '__globals__'):
            data_dict.pop(elem)  # pop the extra system info key-value pairs
        return data_dict


def test_load_data():
    _load_data("kalman_inputs.mat")


if __name__ == "__main__":
    test_load_data()
    print('Test data loading finished without exception.')
